





def add_inverse_graph(graph_values):
    graphs = []
    for graph in graph_values:
        if not graph.startswith("http"):
            graph = '$(' + graph + ')'
        else:
            if '{' in graph:
                graph = graph.replace('{', "$(").replace('}', ')')
        graphs.append(graph.toPython())

    if len(graphs) == 1:
        yarrrml_graphs = {'graph': graphs[0]}
    else:
        yarrrml_graphs = {'graphs': graphs}

    return yarrrml_graphs



import rdflib
from .constants import *
from .graph import add_inverse_graph
from .source import add_source, add_table
from .subject import add_subject
from .termmap import generate_rml_termmap, generate_cc_termmap, generate_rml_termmap_text
from .mapping import prefixes
from ruamel.yaml import YAML


def add_predicate_object_maps(data, mapping, mapping_format):
    po_template = ""
    mapping_data = data.get(YARRRML_MAPPINGS).get(mapping)
    if YARRRML_PREDICATEOBJECT in mapping_data:
        pom_text = "\t" + R2RML_PREDICATE_OBJECT_MAP + " [\n"
        for predicate_object_map in mapping_data.get(YARRRML_PREDICATEOBJECT):
            po_template += pom_text + add_predicate_object(data, mapping, predicate_object_map, mapping_format) + "\n"
    else:
        logger.warning("The triples map " + mapping + " does not have predicate object maps defined")
    return po_template


def add_predicate_object(data, mapping, predicate_object, mapping_format=RML_URI):
    template = ""

    predicate_list = []
    predicate_list.extend(predicate_object.get(YARRRML_PREDICATES))
    for pm in predicate_list:
        pm_value = pm
        execution = False
        if YARRRML_VALUE in pm and type(pm) is dict:
            pm_value = pm[YARRRML_VALUE]
        elif YARRRML_FUNCTION in pm:
            pm_value = pm[YARRRML_FUNCTION]
            execution = True
        template += generate_rml_termmap(R2RML_PREDICATE, R2RML_PREDICATE_CLASS, pm_value, "\t\t\t")
        if execution:
            template = template.replace(R2RML_CONSTANT + " \"" + pm_value + "\"", RML_EXECUTION + " <" + pm_value + ">")
        if YARRRML_TARGETS in pm:
            template = template[0:-3] + "\t" + RML_LOGICAL_TARGET + " <" + pm[YARRRML_TARGETS] + ">\n\t\t];\n"

    object_maps = predicate_object.get(YARRRML_OBJECTS)
    for om in object_maps:
        object_value = om.get(YARRRML_VALUE)
        if YARRRML_GATHER in om and YARRRML_GATHER_AS in om:
            gather=om[YARRRML_GATHER]
            if isinstance(gather, list):
                gather = gather[0]
            if YARRRML_MAPPING in gather or YARRRML_CONDITION in gather:
                template += ref_cc_mapping(data, mapping, gather, YARRRML_MAPPING, R2RML_PARENT_TRIPLESMAP, mapping_format)
                if YARRRML_GATHER_AS in om and om[YARRRML_GATHER_AS] in YARRRML_GATHER_AS_OPTIONS:
                    template += "\t\t\t" + RML_CC_GATHER_AS + " rdf:" + om[YARRRML_GATHER_AS].capitalize() + ";\n"
                if YARRRML_VALUE in om:
                    text = om.get('value', '')
                    term_map, text = generate_rml_termmap_text(text, mapping_format)
                    if term_map == STAR_QUOTED:
                        if 'quoted' in text:
                            template += "\t\t\t" + term_map + " <" + text[YARRRML_QUOTED] + "_0>;\n"
                        else:
                            template += "\t\t\t" +  term_map + " <" + text[YARRRML_NON_ASSERTED] + "_0>;\n"
                    elif term_map != "rr:constant":
                        template += "\t\t\t" +  term_map + " \"" + text + "\";\n"
                    else:
                        if text.startswith("http"):
                            template += "\t\t\t" +  term_map + " <" + text + ">;\n"
                        else:
                            if ":" in text or "<" in text:
                                template +=  "\t\t\t" + term_map + " " + text + ";\n"
                            else:
                                template += "\t\t\t" +  term_map + " \"" + text + "\";\n"
                template += "\n\t\t];\n"
            else:
                template += generate_cc_termmap(STAR_OBJECT, R2RML_OBJECT_CLASS, om, "\t\t\t", mapping_format) + "\n\t\t];\n"

            if YARRRML_TYPE in om:
                if om.get(YARRRML_TYPE) == 'iri':
                    template = template[0:len(template) - 6] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_IRI + "\n\t\t];\n"
                elif om.get(YARRRML_TYPE) == 'blank':
                    template = template[0:len(template) - 6] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_BLANK_NODE + "\n\t\t];\n"
                elif om.get(YARRRML_TYPE) == 'literal':
                    template = template[0:len(template) - 6] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_LITERAL + "\n\t\t];\n"
        elif object_value is not None:
            rml_map_class, rml_map, r2rml_map = None, None, None
            if mapping_format == STAR_URI:
                rml_property = STAR_OBJECT
            else:
                rml_property = R2RML_OBJECT
            template += generate_rml_termmap(rml_property, R2RML_OBJECT_CLASS, object_value, "\t\t\t", mapping_format)

            optional_value = None
            if YARRRML_DATATYPE in om:
                optional_value = om[YARRRML_DATATYPE]
                rml_map = RML_DATATYPE_MAP
                rml_map_class = RML_DATATYPE_MAP_CLASS
                r2rml_map = R2RML_DATATYPE
            if YARRRML_LANGUAGE in om:
                if '$(' in om[YARRRML_LANGUAGE]:
                    optional_value = om[YARRRML_LANGUAGE]
                else:
                    optional_value = '"' + om[YARRRML_LANGUAGE] + '"'
                rml_map = RML_LANGUAGE_MAP
                rml_map_class = RML_LANGUAGE_MAP_CLASS
                r2rml_map = R2RML_LANGUAGE
            if optional_value is not None:
                if "$(" in optional_value:
                    template = template[0:len(template) - 5] + generate_rml_termmap(rml_map,
                                                                                    rml_map_class,
                                                                                    optional_value,
                                                                                    "\t\t\t\t",
                                                                                    mapping_format) + "\t\t];\n"

                else:
                    template = template[0:len(template) - 5] + "\t\t\t" + r2rml_map + " " + optional_value + "\n\t\t];\n"

            elif YARRRML_TARGETS in om:
                template = template[0:len(template) - 5] + "\t\t\t" + RML_LOGICAL_TARGET + " <" + om[YARRRML_TARGETS] + ">\n\t\t];\n"
            if YARRRML_TYPE in om:
                if om.get(YARRRML_TYPE) == 'iri':
                    template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_IRI + "\n\t\t];\n"
                elif om.get(YARRRML_TYPE) == 'blank':
                    template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_BLANK_NODE + "\n\t\t];\n"
                elif om.get(YARRRML_TYPE) == 'literal':
                    template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " + R2RML_LITERAL + "\n\t\t];\n"

        elif YARRRML_MAPPING in om or YARRRML_NON_ASSERTED in om or YARRRML_QUOTED in om:
            if YARRRML_MAPPING in om:
                template += ref_mapping(data, mapping, om, YARRRML_MAPPING, R2RML_PARENT_TRIPLESMAP, mapping_format)
            elif YARRRML_NON_ASSERTED in om:
                if YARRRML_CONDITION in om:
                    template += ref_mapping(data, mapping, om, YARRRML_NON_ASSERTED, STAR_QUOTED, mapping_format)
                else:
                    template += generate_rml_termmap(STAR_OBJECT, STAR_CLASS, om, "\t\t\t", mapping_format)
            else:
                template += ref_mapping(data, mapping, om, YARRRML_QUOTED, STAR_QUOTED, mapping_format)

        else:
            object_value = om

            if mapping_format == STAR_URI:
                template += generate_rml_termmap(STAR_OBJECT, R2RML_OBJECT_CLASS,
                                                 object_value, "\t\t\t", mapping_format)
            elif YARRRML_FUNCTION in om:
                template += generate_rml_termmap(R2RML_OBJECT, R2RML_OBJECT_CLASS, om[YARRRML_FUNCTION], "\t\t\t",
                                                 mapping_format)
                template = template.replace(R2RML_CONSTANT + " \"" + om[YARRRML_FUNCTION] + "\"",
                                            RML_EXECUTION + " <" + om.get(YARRRML_FUNCTION) + ">")
            else:
                template += generate_rml_termmap(R2RML_OBJECT, R2RML_OBJECT_CLASS,
                                                 object_value, "\t\t\t", mapping_format)
            if type(om) is dict:
                if YARRRML_DATATYPE in om:
                    template = template[0:len(template) - 5] + "\t\t\t" + R2RML_DATATYPE + " " \
                               + om.get(YARRRML_DATATYPE) + "\n\t\t];\n"
                if YARRRML_LANGUAGE in om:
                    template = template[0:len(template) - 5] + "\t\t\t" + R2RML_LANGUAGE + " \"" \
                               + om.get(YARRRML_LANGUAGE) + "\"\n\t\t];\n"
                if YARRRML_TYPE in om:
                    if om.get(YARRRML_TYPE) == "iri":
                        iri = True
                    elif om.get(YARRRML_TYPE) == "literal":
                        template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " \
                                   + R2RML_LITERAL + "\n\t\t];\n"
                    elif om.get(YARRRML_TYPE) == YARRRML_BLANK:
                        template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " \
                                   + R2RML_BLANK_NODE + "\n\t\t];\n"
                if YARRRML_TARGETS in om:
                    template = template[0:len(template) - 5] + "\t\t\t" + RML_LOGICAL_TARGET + " <" + om.get(
                        YARRRML_TARGETS) + ">\n\t\t];\n"

            if YARRRML_IRI in om:
                template = template[0:len(template) - 5] + "\t\t\t" + R2RML_TERMTYPE + " " \
                           + R2RML_IRI + "\n\t\t];\n"

    if YARRRML_GRAPHS in predicate_object:
        graph_list = predicate_object.get(YARRRML_GRAPHS)
        for graph in graph_list:
            graph_value = graph
            if YARRRML_VALUE in graph:
                graph_value = graph[YARRRML_VALUE]
            template += generate_rml_termmap(R2RML_GRAPH_MAP, R2RML_GRAPH_CLASS, graph_value, "\t\t\t")
            if YARRRML_TARGETS in graph:
                template = template[0:-3] + "\t" + RML_LOGICAL_TARGET + " <" + graph[
                    YARRRML_TARGETS] + ">\n\t\t];\n"

    return template + "\t];"


def ref_mapping(data, mapping, om, yarrrml_key, ref_type_property, mapping_format):
    list_mappings = []
    template = ""
    object = R2RML_OBJECT
    for mappings in data.get(YARRRML_MAPPINGS):
        list_mappings.append(mappings)

    mapping_join = om.get(yarrrml_key)

    if mapping_join in list_mappings:
        subject_list = add_subject(data, mapping_join, mapping_format)
        if mapping_format == R2RML_URI:
            source_list = add_table(data, mapping_join)
        else:
            if mapping_format == STAR_URI:
                object = STAR_OBJECT
            source_list, external_references = add_source(data, mapping_join)

        number_joins_rml = len(subject_list) * len(source_list)
        for i in range(number_joins_rml):
            template += "\t\t" + object + \
                        " [\n\t\t\ta " + R2RML_REFOBJECT_CLASS + \
                        ";\n\t\t\t" + ref_type_property + " <" + mapping_join + "_" + str(i) + ">;\n"
            if YARRRML_CONDITION in om:
                conditions = om.get(YARRRML_CONDITION)
                if type(conditions) is not list:
                    conditions = [conditions]
                for condition in conditions:
                    if YARRRML_PARAMETERS in condition:
                        list_parameters = condition.get(YARRRML_PARAMETERS)
                        if len(list_parameters) == 2:

                            try:
                                child = list_parameters[0]['value'].replace('"', r'\"').replace("$(", '"').replace(")",
                                                                                                                   '"')
                                parent = list_parameters[1]['value'].replace('"', r'\"').replace("$(", '"').replace(")",
                                                                                                                    '"')

                            except Exception as e:
                                logger.error("ERROR: Parameters not normalized correctly")

                            template += "\t\t\t" + R2RML_JOIN_CONITION + \
                                        " [\n\t\t\t\t" + R2RML_CHILD + " " + child + \
                                        ";\n\t\t\t\t" + R2RML_PARENT + " " + parent + ";\n\t\t\t];\n"

                        else:
                            logger.error("Error in reference mapping another mapping in mapping " + mapping)
                            raise Exception("Only two parameters can be indicated (child and parent)")
                template += "\t\t];\n"
            else:
                template += "\n\t\t]\n"

    else:
        logger.error("Error in reference another mapping in mapping " + mapping)
        raise Exception("Review how is defined the reference to other mappings")

    return template

def ref_cc_mapping(data, mapping, om, yarrrml_key, ref_type_property, mapping_format):
    list_mappings = []
    template = ""
    object = R2RML_OBJECT
    for mappings in data.get(YARRRML_MAPPINGS):
        list_mappings.append(mappings)

    mapping_join = om.get(yarrrml_key)

    if mapping_join in list_mappings:
        subject_list = add_subject(data, mapping_join, mapping_format)
        if mapping_format == R2RML_URI:
            source_list = add_table(data, mapping_join)
        else:
            if mapping_format == STAR_URI:
                object = STAR_OBJECT
            source_list, external_references = add_source(data, mapping_join)

        number_joins_rml = len(subject_list) * len(source_list)
        for i in range(number_joins_rml):
            template += "\t\t" + object + \
                        " [\n\t\t\ta " + R2RML_OBJECT_CLASS + ";\n\t\t\t" + RML_CC_GATHER + \
                        " (\n\t\t\t\t[\n\t\t\t\t\t" + ref_type_property + " <" + mapping_join + "_" + str(i) + ">;\n"
            if YARRRML_CONDITION in om:
                conditions = om.get(YARRRML_CONDITION)
                if type(conditions) is not list:
                    conditions = [conditions]
                for condition in conditions:
                    if YARRRML_PARAMETERS in condition:
                        list_parameters = condition.get(YARRRML_PARAMETERS)
                        if len(list_parameters) == 2:

                            try:
                                child = list_parameters[0]['value'].replace('"', r'\"').replace("$(", '"').replace(")",
                                                                                                                   '"')
                                parent = list_parameters[1]['value'].replace('"', r'\"').replace("$(", '"').replace(")",
                                                                                                                    '"')

                            except Exception as e:
                                logger.error("ERROR: Parameters not normalized correctly")

                            template += "\t\t\t\t\t" + R2RML_JOIN_CONITION + \
                                        " [\n\t\t\t\t\t\t" + R2RML_CHILD + " " + child + \
                                        ";\n\t\t\t\t\t\t" + R2RML_PARENT + " " + parent + ";\n\t\t\t\t\t];\n"

                        else:
                            logger.error("Error in reference mapping another mapping in mapping " + mapping)
                            raise Exception("Only two parameters can be indicated (child and parent)")
                template += "\t\t\t\t]\n\t\t\t);\n"
            else:
                template += "\t\t\t\t]\n\t\t\t);\n"

    else:
        logger.error("Error in reference another mapping in mapping " + mapping)
        raise Exception("Review how is defined the reference to other mappings")


    return template


def add_inverse_pom(mapping_id, rdf_mapping, classes, prefixes):
    yarrrml_poms = []
    yaml = YAML()
    for c in classes:
        yarrrml_pom = yaml.seq(['rdf:type', find_prefixes(c.toPython(), prefixes)])
        yarrrml_pom.fa.set_flow_style()
        yarrrml_poms.append(yarrrml_pom)

    query = f'SELECT ?predicate ?predicateValue ?object ?objectValue ?termtype ?datatype ?datatypeMapValue ' \
            f'?language ?languageMapValue ?parentTriplesMap ?child ?parent ?graphValue' \
            f' WHERE {{ ' \
            f'<{mapping_id}> {R2RML_PREDICATE_OBJECT_MAP} ?predicateObjectMap . ' \
            f'?predicateObjectMap {R2RML_PREDICATE}|{R2RML_SHORTCUT_PREDICATE} ?predicate .' \
            f'OPTIONAL {{ ?predicate {R2RML_TEMPLATE}|{R2RML_COLUMN}|{R2RML_CONSTANT}|{RML_REFERENCE} ?predicateValue . }}' \
            f'?predicateObjectMap {R2RML_OBJECT}|{R2RML_SHORTCUT_OBJECT} ?object .' \
            f' {{ OPTIONAL {{ ?predicateObjectMap {R2RML_GRAPH} ?graphValue .}}' \
            f' }} UNION {{' \
            f' OPTIONAL {{ ' \
            f' ?predicateObjectMap {R2RML_GRAPH_MAP} ?graphMap . ' \
            f' ?graphMap {R2RML_TEMPLATE}|{R2RML_CONSTANT}|{RML_REFERENCE} ?graphValue .}} }}' \
            f'OPTIONAL {{ ' \
            f'?object {R2RML_TEMPLATE}|{R2RML_COLUMN}|{R2RML_CONSTANT}|{RML_REFERENCE} ?objectValue .' \
            f'OPTIONAL {{ ?object {R2RML_TERMTYPE} ?termtype . }}' \
            f'OPTIONAL {{ ?object {R2RML_DATATYPE} ?datatype .}} ' \
            f'OPTIONAL {{ ' \
            f' ?object {RML_DATATYPE_MAP} ?datatypeMap .' \
            f' ?datatypeMap {R2RML_TEMPLATE}|{R2RML_CONSTANT}|{RML_REFERENCE} ?datatypeMapValue .}} ' \
            f'OPTIONAL {{ ?object {R2RML_LANGUAGE} ?language .}} ' \
            f'OPTIONAL {{ ' \
            f' ?object {RML_LANGUAGE_MAP} ?languageMap .' \
            f' ?languageMap {R2RML_TEMPLATE}|{R2RML_CONSTANT}|{RML_REFERENCE} ?languageMapValue .}} }} ' \
            f'OPTIONAL {{ ?object {R2RML_PARENT_TRIPLESMAP} ?parentTriplesMap .' \
            f'OPTIONAL {{ ' \
            f'?object {R2RML_JOIN_CONITION} ?join_condition .' \
            f'?join_condition {R2RML_CHILD} ?child .' \
            f'?join_condition {R2RML_PARENT} ?parent }} }}' \
            f'}}'

    for tm in rdf_mapping.query(query):
        yarrrml_pom = []
        if tm['predicateValue']:
            predicate = tm['predicateValue'].toPython()
        elif tm['predicate']:
            predicate = tm['predicate'].toPython()
        else:
            logger.error("ERROR: There is POM without predicate map defined")
            raise Exception("Review your mapping " + str(mapping_id))

        if not predicate.startswith("http"):
            predicate = '$(' + predicate + ')'
        elif predicate.startswith("http") and "{" not in predicate:
            predicate = find_prefixes(predicate, prefixes)
        else:
            predicate = predicate.replace('{', '$(').replace('}', ')')

        predicate = find_prefixes(predicate, prefixes)

        if tm['parentTriplesMap']:
            if tm['child']:
                yarrrml_pom = {'p': predicate, 'o': {'mapping': None, 'condition':
                    {'function': 'equal', 'parameters': []}}}
                yarrrml_pom['o']['mapping'] = tm['parentTriplesMap'].split("/")[-1]
                child = yaml.seq(['str1', '$(' + tm['child'] + ')'])
                child.fa.set_flow_style()
                parent = yaml.seq(['str2', '$(' + tm['parent'] + ')'])
                parent.fa.set_flow_style()
                yarrrml_pom['o']['condition']['parameters'].append(child)
                yarrrml_pom['o']['condition']['parameters'].append(parent)
            else:
                yarrrml_pom = {'p': predicate, 'o': {'mapping': tm['parentTriplesMap'].split("/")[-1]}}


        else:
            datatype = None
            language = None

            if tm['objectValue']:  # we have extended objectMap version
                object = tm['objectValue'].toPython()
            elif tm['object']:
                object = tm['object'].toPython()
            else:
                logger.error("There is not object for a given predicate")
                raise Exception("Review your mapping " + str(mapping_id))

            if not object.startswith("http"):
                object = '$(' + object + ')'
            elif object.startswith("http") and "{" not in object:
                object = find_prefixes(object, prefixes)
            else:
                object = object.replace('{', '$(').replace('}', ')')

            if tm['termtype']:
                if tm['termtype'] == rdflib.URIRef(R2RML_IRI):
                    object = object + '~iri'

            if tm['graphValue']:
                graph_value = add_inverse_graph([tm['graphValue']])
                yarrrml_pom = {'p': predicate, 'o': object}
                yarrrml_pom.update(graph_value)
            else:
                yarrrml_pom.append(predicate)
                yarrrml_pom.append(object)

            if tm[YARRRML_DATATYPE]:
                datatype = tm[YARRRML_DATATYPE].toPython()
                prefix = list({i for i in prefixes if datatype.startswith(prefixes[i])})
                if prefix:
                    datatype = datatype.replace(prefixes[prefix[0]], prefix[0] + ":")
            elif tm['datatypeMapValue']:
                datatype = tm['datatypeMapValue']
                if not datatype.startswith("http"):
                    datatype = '$(' + datatype + ')'
                else:
                    datatype.replace('{', '$(').replace('}', ')')
            if tm[YARRRML_LANGUAGE]:
                language = tm[YARRRML_LANGUAGE].toPython() + "~lang"
            elif tm['languageMapValue']:
                language = tm['languageMapValue']
                if not language.startswith("http"):
                    language = '$(' + language + ')'
                else:
                    language.replace('{', '$(').replace('}', ')')

            if type(yarrrml_pom) is list:
                if datatype:
                    datatype = find_prefixes(datatype, prefixes)
                    yarrrml_pom.append(datatype)
                if language:
                    yarrrml_pom.append(language)
            elif type(yarrrml_pom) is dict:
                if datatype:
                    datatype = find_prefixes(datatype, prefixes)
                    yarrrml_pom[YARRRML_DATATYPE] = datatype
                if language:
                    yarrrml_pom[YARRRML_LANGUAGE] = language

        if type(yarrrml_pom) is list:
            yarrrml_pom = yaml.seq(yarrrml_pom)
            yarrrml_pom.fa.set_flow_style()
        yarrrml_poms.append(yarrrml_pom)

    return yarrrml_poms


def find_prefixes(text, prefixes):
    prefix = list({i for i in prefixes if text.startswith(prefixes[i])})
    if len(prefix) > 0:
        text = text.replace(prefixes[prefix[0]], prefix[0] + ":")
    return text



import logging, coloredlogs

##############################################################################
#############################   RML CONSTANTS  ###############################
##############################################################################

RML_URI = 'http://semweb.mmlab.be/ns/rml#'
R2RML_URI = 'http://www.w3.org/ns/r2rml#'
RDF_URI = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
D2RQ_URI = 'http://www.wiwiss.fu-berlin.de/suhl/bizer/D2RQ/0.1#'
QL_URI = 'http://semweb.mmlab.be/ns/ql#'
EXAMPLE_URI = 'http://example.com/ns#'
RDFS_URI = 'http://www.w3.org/2000/01/rdf-schema#'
XSD_URI = 'http://www.w3.org/2001/XMLSchema#'
FOAF_URI ='http://xmlns.com/foaf/0.1/'
RDF_TYPE = 'rdf:type'
SCHEMA_URI = 'http://schema.org/'
STAR_URI = 'https://w3id.org/kg-construct/rml-star'
COMPRESSION_URI = 'http://semweb.mmlab.be/ns/rml-compression#'
FORMATS_URI = 'http://www.w3.org/ns/formats/'
VOID_URI = 'http://rdfs.org/ns/void#'
FNML_URI = 'http://semweb.mmlab.be/ns/fnml#'
GREL_URI = 'http://users.ugent.be/~bjdmeest/function/grel.ttl#'
SD_URI = 'https://w3id.org/okn/o/sd#'

RML_PREFIX = '@prefix'
RML_BASE = '@base'
RML_LOGICAL_SOURCE_CLASS = 'rml:LogicalSource'
RML_LOGICAL_SOURCE = 'rml:logicalSource'
RML_SOURCE = 'rml:source'
RML_REFERENCE_FORMULATION = 'rml:referenceFormulation'
RML_REFERENCE_FORMULATION_CLASS = 'rml:ReferenceFormulation'
RML_ITERATOR = 'rml:iterator'
RML_REFERENCE = 'rml:reference'
RML_LANGUAGE_MAP = 'rml:languageMap'
RML_LANGUAGE_MAP_CLASS = 'rml:LanguageMap'
RML_DATATYPE_MAP = 'rml:datatypeMap'
RML_DATATYPE_MAP_CLASS = 'rml:DatatypeMap'
RML_QUERY = 'rml:query'

RML_LOGICAL_TARGET = 'rml:logicalTarget'
RML_LOGICAL_TARGET_CLASS = 'rml:LogicalTarget'
RML_TARGET = 'rml:target'
RML_SERIALIZATION = 'rml:serialization'
RML_COMPRESSION = 'rml:compression'

RML_EXECUTION = 'fnml:execution'
RML_EXECUTION_CLASS = 'fnml:Execution'
RML_RETURN = 'fnml:return'
RML_FUNCTION = 'fnml:function'
RML_INPUT = 'fnml:input'
RML_INPUT_CLASS = 'fnml:Input'
RML_PARAMETER = 'fnml:parameter'
RML_VALUE_MAP = 'fnml:valueMap'
RML_VALUE_MAP_CLASS = 'fnml:ValueMap'

RML_CC_GATHER = 'rml:gather'
RML_CC_GATHER_AS = 'rml:gatherAs'
RML_CC_EMPTY = 'rml:allowEmptyListAndContainer'
RML_CC_STRATEGY = 'rml:strategy'

STAR_CLASS = 'rml:StarMap'
STAR_NON_ASSERTED_CLASS = 'rml:NonAssertedTriplesMap'
STAR_QUOTED = 'rml:quotedTriplesMap'
STAR_SUBJECT = 'rml:subjectMap'
STAR_OBJECT = 'rml:objectMap'

R2RML_TEMPLATE = 'rr:template'
R2RML_TRIPLES_MAP = 'rr:TriplesMap'
R2RML_CONSTANT = 'rr:constant'
R2RML_SUBJECT = 'rr:subjectMap'
R2RML_SUBJECT_CLASS = 'rr:SubjectMap'
R2RML_CLASS = 'rr:class'
R2RML_SQL_VERSION = 'rr:sqlVersion'
R2RML_SQL_QUERY = 'rr:sqlQuery'
R2RML_PREDICATE_OBJECT_MAP = 'rr:predicateObjectMap'
R2RML_PREDICATE_OBJECT_MAP_CLASS = 'rr:PredicateObjectMap'
R2RML_SHORTCUT_PREDICATE = 'rr:predicate'
R2RML_PREDICATE = 'rr:predicateMap'
R2RML_PREDICATE_CLASS = 'rr:PredicateMap'
R2RML_SHORTCUT_OBJECT = 'rr:object'
R2RML_OBJECT = 'rr:objectMap'
R2RML_OBJECT_CLASS = 'rr:ObjectMap'
R2RML_GRAPH = 'rr:graph'
R2RML_GRAPH_MAP = 'rr:graphMap'
R2RML_GRAPH_CLASS = 'rr:GraphMap'
R2RML_DATATYPE = 'rr:datatype'
R2RML_TERMTYPE = 'rr:termType'
R2RML_LANGUAGE = 'rr:language'
R2RML_IRI = 'rr:IRI'
R2RML_BLANK_NODE = 'rr:BlankNode'
R2RML_LITERAL = 'rr:Literal'
R2RML_REFOBJECT_CLASS = 'rr:RefObjectMap'
R2RML_PARENT_TRIPLESMAP = 'rr:parentTriplesMap'
R2RML_JOIN_CONITION = 'rr:joinCondition'
R2RML_CHILD = 'rr:child'
R2RML_PARENT = 'rr:parent'
R2RML_LOGICAL_TABLE_CLASS = 'rr:LogicalTable'
R2RML_LOGICAL_TABLE = 'rr:logicalTable'
R2RML_TABLE_NAME = 'rr:tableName'
R2RML_COLUMN = 'rr:column'


##############################################################################
#############################   D2RQ CONSTANTS  ###########################
##############################################################################
D2RQ_DATABASE_CLASS = 'd2rq:Database'
D2RQ_DSN = 'd2rq:jdbcDSN'
D2RQ_DRIVER = 'd2rq:jdbcDriver'
D2RQ_USER = 'd2rq:username'
D2RQ_PASS = 'd2rq:password'

##############################################################################
#############################   SD CONSTANTS  ###########################
##############################################################################
SD_DATASET_SPEC = 'sd:DatasetSpecification'
SD_NAME = 'sd:name'
SD_HAS_DATA_TRANSFORMATION = 'sd:hasDataTransformation'
SD_HAS_SOFTWARE_REQUIREMENTS = 'sd:hasSoftwareRequirements'
SD_HAS_SOURCE_CODE= 'sd:hasSourceCode'
SD_PROGRAMMING_LANGUAGE =  'sd:programmingLanguage'
KG4DI_DEFINED_BY = 'kg4di:definedBy'
##############################################################################
#############################   YARRRML CONSTANTS  ###########################
##############################################################################

YARRRML_PREFIXES = 'prefixes'
YARRRML_SOURCES = 'sources'
YARRRML_TABLE = 'table'
YARRRML_ACCESS = 'access'
YARRRML_QUERY = 'query'
YARRRML_REFERENCE_FORMULATION = 'referenceFormulation'
YARRRML_QUERY_FORMULATION = 'queryFormulation'
YARRRML_ITERATOR = 'iterator'
YARRRML_CREDENTIALS = 'credentials'
YARRRML_TYPE = 'type'
YARRRML_USERNAME = 'username'
YARRRML_PASSWORD = 'password'

YARRRML_STRUCTURE_DEFINER = 'structureDefiner'
YARRRML_SOFTWARE_SPECIFICATION = 'softwareSpecification'
YARRRML_PROGRAMMING_LANGUAGE = 'programmingLanguage'
YARRRML_SOFTWARE_REQUIREMENTS = 'softwareRequirements'

YARRRML_MAPPINGS = 'mappings' # used for mappings in conditions and mappings main key
YARRRML_MAPPING = 'mapping'

YARRRML_SUBJECTS = 'subjects'
YARRRML_AUTHORS = 'authors'
YARRRML_GRAPHS = 'graphs'

YARRRML_PREDICATEOBJECT = 'predicateobjects'

YARRRML_PREDICATES = 'predicates'
YARRRML_OBJECTS = 'objects'
YARRRML_VALUE = 'value'
YARRRML_DATATYPE = 'datatype'
YARRRML_LANGUAGE = 'language'

YARRRML_CONDITION = 'condition'
YARRRML_EQUAL = 'equal'
YARRRML_JOIN = 'join'
YARRRML_PARAMETERS = 'parameters' #used for conditions and functions


YARRRML_IRI = '~iri'
YARRRML_LANG = '~lang'
YARRRML_BLANK = 'blank'

YARRRML_QUOTED = 'quoted'
YARRRML_NON_ASSERTED = 'quotedNonAsserted'

YARRRML_TARGETS = 'targets'
YARRRML_SERIALIZATION = 'serialization'
YARRRML_COMPRESSION = 'compression'

YARRRML_FUNCTION = 'function'

YARRRML_PARAMETER = 'parameter'

YARRRML_GATHER = 'gather'
YARRRML_GATHER_AS = 'gatherAs'
YARRRML_GATHER_AS_OPTIONS = ["alt","bag","seq","list"]
YARRRML_EMPTY = 'empty'
YARRRML_STRATEGY = 'strategy'

YARRRML_MAPPING_KEYS = [YARRRML_MAPPINGS, YARRRML_MAPPING]
YARRRML_SUBJECT_KEYS = [YARRRML_SUBJECTS]
YARRRML_POM_KEYS = [YARRRML_PREDICATEOBJECT]
YARRRML_GRAPH_KEYS = [YARRRML_GRAPHS]
YARRRML_PREDICATE_KEYS = [YARRRML_PREDICATES]
YARRRML_OBJECT_KEYS = [YARRRML_OBJECTS]
YARRRML_FUNCTION_KEYS = [YARRRML_FUNCTION]
YARRRML_PARAMETERS_KEYS = [YARRRML_PARAMETERS]
YARRRML_PARAMETER_KEYS = [YARRRML_PARAMETER]
YARRRML_VALUE_KEYS = [YARRRML_VALUE]


YARRRML_OUTPUT_FORMAT = {
    'jsonld':'JSON-LD',
    'n3':'N3',
    'ntriples':'N-Triples',
    'nquads':'N-Quads',
    'ldpatch':'LD_Patch',
    'microdata':'microdata',
    'owlxml':'OWL_XML',
    'owlfunctional':'OWL_Functional',
    'oxmlmanchester':'OWL_Manchester',
    'powder':'POWDER',
    'powder-s':'POWDER-S',
    'prov-n':'PROV-N',
    'prov-xml':'PROV-XML',
    'rdfa':'RDFa',
    'rdfjson':'RDF_JSON',
    'rdfxml':'RDF_XML',
    'rifxml':'RIF_XML',
    'sparqlxml':'SPARQL_Results_XML',
    'sparqljson':'SPARQL_Results_JSON',
    'sparqlcsv':'SPARQL_Results_CSV',
    'sparqltsv':'SPARQL_Results_TSV',
    'turtle':'Turtle',
    'trig':'TriG',
}

YARRRML_REFERENCE_FORMULATIONS = {
    'csv': 'CSV',
    'json': 'JSONPath',
    'xpath': 'XPath',
    'jsonpath': 'JSONPath',
    "shp": "SHP",
    "xlsx": "CSV"
}

YARRRML_DATABASES_DRIVER = {
    'mysql': 'com.mysql.jdbc.Driver',
    'postgresql': 'org.postgresql.Driver',
    'sqlserver': 'com.microsoft.sqlserver.jdbc.SQLServerDriver'
}

logger = logging.getLogger(__name__)
coloredlogs.install(level='DEBUG', fmt='%(asctime)s,%(msecs)03d | %(levelname)s: %(message)s', logger=logger)



from . import *

prefixes = {}


def add_mapping(mapping, mappings, it):
    map_template = "<" + mapping + "_" + str(it) + "> a "
    if mappings[mapping]:
        if mappings[mapping] == "non_asserted":
            map_template += STAR_NON_ASSERTED_CLASS + ", "
    map_template += R2RML_TRIPLES_MAP + ";\n\n"
    return map_template


def add_prefix(data):
    global prefixes
    template = []
    common_prefixes = []
    if YARRRML_PREFIXES in data:
        prefixes = data.get(YARRRML_PREFIXES)
        for prefix in prefixes:
            prefix_uri = data.get(YARRRML_PREFIXES).get(prefix)
            check_common_prefixes(prefix_uri, common_prefixes)
            template.append(RML_PREFIX + " " + prefix + ": <" + prefix_uri + ">.\n")

    if "r2rml" not in common_prefixes:
        template.append(RML_PREFIX + " rr: <" + R2RML_URI + ">.\n")
    if "rml" not in common_prefixes:
        template.append(RML_PREFIX + " rml: <" + RML_URI + ">.\n")
    if "rdf" not in common_prefixes:
        template.append(RML_PREFIX + " rdf: <" + RDF_URI + ">.\n")
    if "rdfs" not in common_prefixes:
        template.append(RML_PREFIX + " rdfs: <" + RDFS_URI + ">.\n")
    if "xsd" not in common_prefixes:
        template.append(RML_PREFIX + " xsd: <" + XSD_URI + ">.\n")
    if "ql" not in common_prefixes:
        template.append(RML_PREFIX + " ql: <" + QL_URI + ">.\n")
    if "d2rq" not in common_prefixes:
        template.append(RML_PREFIX + " d2rq: <" + D2RQ_URI + ">.\n")
    if "foaf" not in common_prefixes:
        template.append(RML_PREFIX + " foaf: <" + FOAF_URI + ">.\n")
    if "schema" not in common_prefixes:
        template.append(RML_PREFIX + " schema: <" + SCHEMA_URI + ">.\n")
    if "formats" not in common_prefixes:
        template.append(RML_PREFIX + " formats: <" + FORMATS_URI + ">.\n")
    if "comp" not in common_prefixes:
        template.append(RML_PREFIX + " comp: <" + COMPRESSION_URI + ">.\n")
    if "void" not in common_prefixes:
        template.append(RML_PREFIX + " void: <" + VOID_URI + ">.\n")
    if "fnml" not in common_prefixes:
        template.append(RML_PREFIX + " fnml: <" + FNML_URI + ">.\n")
    if "grel" not in common_prefixes:
        template.append(RML_PREFIX + " grel: <" + GREL_URI + ">.\n")
    if "base" not in common_prefixes:
        template.append(RML_BASE + " <" + EXAMPLE_URI + ">.\n")

    template.append("\n\n")

    return "".join(template)


def check_common_prefixes(prefix_uri, common_prefixes):
    if prefix_uri == R2RML_URI:
        common_prefixes.append("r2rml")
    elif prefix_uri == RML_URI:
        common_prefixes.append("rml")
    elif prefix_uri == RDF_URI:
        common_prefixes.append("rdf")
    elif prefix_uri == QL_URI:
        common_prefixes.append("ql")
    elif prefix_uri == D2RQ_URI:
        common_prefixes.append("d2rq")
    elif prefix_uri == RDFS_URI:
        common_prefixes.append("rdfs")
    elif prefix_uri == XSD_URI:
        common_prefixes.append("xsd")
    elif prefix_uri == FOAF_URI:
        common_prefixes.append("foaf")
    elif prefix_uri == SCHEMA_URI:
        common_prefixes.append("schema")
    elif prefix_uri == FORMATS_URI:
        common_prefixes.append("formats")
    elif prefix_uri == COMPRESSION_URI:
        common_prefixes.append("comp")
    elif prefix_uri == VOID_URI:
        common_prefixes.append("void")
    elif prefix_uri == FNML_URI:
        common_prefixes.append("fnml")
    elif prefix_uri == GREL_URI:
        common_prefixes.append("grel")


def add_inverse_prefix(rdf_mapping):
    prefixes = {}
    for prefix, uri in rdf_mapping.namespaces():
        if prefix:
            prefixes[prefix] = uri.toPython()
    return prefixes


def get_non_asserted_mappings(yarrrml_data, mapping_format):
    mappings = dict.fromkeys(list(yarrrml_data.get(YARRRML_MAPPINGS).keys()))
    for mapping in yarrrml_data.get(YARRRML_MAPPINGS):
        keys = yarrrml_data.get(YARRRML_MAPPINGS).get(mapping).keys()
        for key in keys:
            if type(yarrrml_data.get(YARRRML_MAPPINGS).get(mapping).get(key)) is list:
                for value in yarrrml_data.get(YARRRML_MAPPINGS).get(mapping).get(key):
                    if type(value) is dict:
                        star_data = [value]
                        if YARRRML_OBJECTS in value:
                            star_data = value[YARRRML_OBJECTS]
                        for val in star_data:
                            if YARRRML_NON_ASSERTED in val:
                                mappings[val[YARRRML_NON_ASSERTED]] = "non_asserted"
                                mapping_format = STAR_URI
                            elif YARRRML_QUOTED in val:
                                mapping_format = STAR_URI

    return mappings, mapping_format


def merge_mapping_section_by_key(key, yarrrml_list):
    output = {key: {}}
    for yarrrml_mapping in yarrrml_list:
        output[key] = output[key] | yarrrml_mapping
    return output



import copy
from .constants import *
key_mapping = {
    YARRRML_MAPPINGS: [YARRRML_MAPPING, 'm'],
    YARRRML_SUBJECTS: ['subject', 's'],
    YARRRML_PREDICATES: ['predicate', 'p'],
    'inversepredicates': ['inversepredicate', 'i'],
    YARRRML_OBJECTS: ['object', 'o'],
    YARRRML_PREDICATEOBJECT: ['po'],
    YARRRML_FUNCTION: ['fn', 'f'],
    YARRRML_PARAMETERS: ['pms'],
    YARRRML_PARAMETER: ['pm'],
    YARRRML_VALUE: ['v'],
    YARRRML_AUTHORS: ['author', 'a'],
    YARRRML_TARGETS: ['target', 't'],
    YARRRML_GRAPHS: ['graph', 'g'],
    YARRRML_SOURCES: ['source', 'src']
}


def normalize_yaml(data):
    if isinstance(data, dict):
        new_data = dict()
        for key, value in data.items():
            new_key = get_normalized_key(key)
            if new_key == YARRRML_AUTHORS and isinstance(value, list):
                new_data[new_key] = expand_authors(value)
            elif new_key == YARRRML_SOURCES:
                new_data[new_key] = expand_sources(value)
            elif new_key == YARRRML_TARGETS:
                new_data[new_key] = expand_targets(value, data.get(YARRRML_TARGETS, {}))
            elif new_key == YARRRML_PREDICATEOBJECT:
                new_data[new_key] = expand_predicateobjects(value)
            elif new_key == YARRRML_SUBJECTS:
                new_data[new_key] = expand_subjects(value, data.get(YARRRML_TARGETS, {}))
            elif new_key == YARRRML_PARAMETERS:
                new_data[new_key] = expand_parameters(value)
            else:
                new_data[new_key] = normalize_yaml(value)


        return new_data

    elif isinstance(data, list):
        new_list = list()
        for item in data:
            new_list.append(normalize_yaml(item))
        return new_list
    return data


def get_normalized_key(key):
    for normalized_key, variants in key_mapping.items():
        if key == normalized_key or key in variants:
            return normalized_key
    return key


def expand_authors(authors):
    expanded_authors = list()
    for author in authors:
        if isinstance(author, str):
            expanded_author = dict()
            parts = author.split()
            name = []
            email = None
            website = None
            for part in parts:
                if isinstance(part, str):
                    if part.startswith("http://") or part.startswith("https://"):
                        expanded_authors.append(part)
                        break
                    elif part.startswith("<") and part.endswith(">"):
                        email = part.strip("<>")
                    elif part.startswith("(") and part.endswith(")"):
                        website = part.strip("()")
                    else:
                        name.append(part)
            if name:
                expanded_author['name'] = " ".join(name)
            if email:
                expanded_author['email'] = email
            if website:
                expanded_author['website'] = website
            if expanded_author:
                expanded_authors.append(expanded_author)
        else:
            expanded_authors.append(author)
    return expanded_authors


def expand_sources(sources):
    def expand_source_item(source):
        if isinstance(source, list):
            if len(source) == 2 and isinstance(source[0], str) and '~' in source[0]:
                access, reference = source[0].split('~')
                if '-' in reference:
                    reference = reference.split('-')
                    return {
                        YARRRML_ACCESS: access,
                        YARRRML_REFERENCE_FORMULATION: reference[1],
                        YARRRML_STRUCTURE_DEFINER: reference[0],
                        YARRRML_ITERATOR: source[1]
                    }
                else:
                    return {
                        YARRRML_ACCESS: access,
                        YARRRML_REFERENCE_FORMULATION: reference,
                        YARRRML_ITERATOR: source[1]
                    }
            elif len(source) == 1 and isinstance(source[0], str) and '~' in source[0]:
                access, reference = source[0].split('~')
                if '-' in reference:
                    reference = reference.split('-')
                    return {
                        YARRRML_ACCESS: access,
                        YARRRML_REFERENCE_FORMULATION: reference[1] ,
                        YARRRML_STRUCTURE_DEFINER: reference[0],
                    }
                else:
                    return {
                        YARRRML_ACCESS: access,
                        YARRRML_REFERENCE_FORMULATION: reference,
                    }
        elif isinstance(source, dict):
            for key, val in source.items():
                if isinstance(val, list) and len(val) == 2 and '~' in val[0]:
                    access, reference = val[0].split('~')
                    return {key: {YARRRML_ACCESS: access, YARRRML_REFERENCE_FORMULATION: reference, YARRRML_ITERATOR: val[1]}}
                else:
                    return normalize_yaml(source)

        return source

    if isinstance(sources, str):
        sources = [sources]
    if isinstance(sources, list):
        return [expand_source_item(src) for src in sources]
    elif isinstance(sources, dict):
        expanded_sources = dict()
        if YARRRML_ACCESS in sources:
            expanded_sources= [sources]
        else:
            for key, src in sources.items():
                expanded_sources[key] = expand_source_item(src)
        return expanded_sources
    return sources


def expand_targets(targets, root_targets={}):
    def expand_target_item(target):
        if isinstance(target, str) and target in root_targets:
            return root_targets[target]
        elif isinstance(target, list) and len(target) >= 1:
            expanded_target = dict()
            access_type = target[0].split('~')
            expanded_target[YARRRML_ACCESS] = access_type[0]
            if len(access_type) > 1:
                expanded_target[YARRRML_TYPE] = access_type[1]
            if len(target) > 1:
                expanded_target[YARRRML_SERIALIZATION] = target[1]
            if len(target) > 2:
                expanded_target[YARRRML_COMPRESSION] = target[2]
            return expanded_target
        elif isinstance(target, dict):
            return normalize_yaml(target)
        return target

    if isinstance(targets, list):
        return list([expand_target_item(t) for t in targets])
    elif isinstance(targets, dict):
        expanded_targets = dict()
        for key, tgt in targets.items():
            expanded_targets[key] = expand_target_item(tgt)
        return expanded_targets
    return targets


def expand_subjects(subjects, root_targets):
    expanded_subjects = list()
    if isinstance(subjects, str):
        expanded_subjects.append(dict({YARRRML_VALUE: subjects}))
    elif isinstance(subjects, list):
        for subject in subjects:
            expanded_subject = dict()
            if isinstance(subject, str):
                if "~" in subject:
                    expanded_subject[YARRRML_GATHER] = subject.split('~')[0]
                    expanded_subject[YARRRML_GATHER_AS] = subject.split('~')[1]
                else:
                    expanded_subject[YARRRML_VALUE] = subject
            elif isinstance(subject, dict):
                if YARRRML_VALUE in subject:
                    expanded_subject[YARRRML_VALUE] = subject.get(YARRRML_VALUE, '')
                if YARRRML_TARGETS in subject:
                    expanded_subject[YARRRML_TARGETS] = expand_targets(subject[YARRRML_TARGETS], root_targets)
                if YARRRML_QUOTED in subject:
                    expanded_subject[YARRRML_QUOTED] = subject.get(YARRRML_QUOTED, '')
                if YARRRML_NON_ASSERTED in subject:
                    expanded_subject[YARRRML_NON_ASSERTED] = subject.get(YARRRML_NON_ASSERTED, '')
                    if YARRRML_CONDITION in subject:
                        expanded_subject[YARRRML_CONDITION] = subject.get(YARRRML_CONDITION, '')
                        if YARRRML_PARAMETERS in subject[YARRRML_CONDITION]:
                            expanded_subject[YARRRML_CONDITION][YARRRML_PARAMETERS] = expand_parameters(expanded_subject[YARRRML_CONDITION].get(YARRRML_PARAMETERS, ''))
                if YARRRML_EMPTY in subject:
                    expanded_subject[YARRRML_EMPTY] = subject.get(YARRRML_EMPTY, '')
                if YARRRML_GATHER in subject:
                    expanded_subject[YARRRML_GATHER] = subject.get(YARRRML_GATHER, '')
                    if YARRRML_GATHER_AS in subject:
                        expanded_subject[YARRRML_GATHER_AS] = subject.get(YARRRML_GATHER_AS, '')
                if YARRRML_STRATEGY in subject:
                    expanded_subject[YARRRML_STRATEGY] = subject.get(YARRRML_STRATEGY, '')
            expanded_subjects.append(expanded_subject)
    elif isinstance(subjects, dict):
        expanded_subjects = [subjects]
    return expanded_subjects


def expand_predicateobjects(predicateobjects):
    expanded_predicateobjects = list()

    for po in predicateobjects:
        if isinstance(po, list):
            expanded_po = dict()
            if len(po) == 3:

                second_value = po[1]
                if '~' in second_value:
                    expanded_po[YARRRML_PREDICATES] = [{YARRRML_VALUE: po[0]}]
                    if "\"" in po[2]:
                        expanded_po[YARRRML_OBJECTS] = [{YARRRML_EMPTY: po[2]}]
                    else:
                        third_value = "\"" + po[2] + "\""
                        expanded_po[YARRRML_OBJECTS] = [{YARRRML_EMPTY: third_value}]
                    if second_value.split('~')[1] in YARRRML_GATHER_AS_OPTIONS:
                        expanded_po[YARRRML_OBJECTS][0][YARRRML_GATHER] = second_value.split('~')[0]
                        expanded_po[YARRRML_OBJECTS][0][YARRRML_GATHER_AS] = second_value.split('~')[1]

                else:
                    expanded_po[YARRRML_PREDICATES] = [{YARRRML_VALUE: po[0]}]
                    expanded_po[YARRRML_OBJECTS] = [{YARRRML_VALUE: po[1]}]

                    third_value = po[2]
                    if '~' in third_value:
                        if third_value.endswith('lang'):
                            expanded_po[YARRRML_OBJECTS][0][YARRRML_LANGUAGE] = third_value.split('~')[0]
                        elif third_value.split('~')[1] in YARRRML_GATHER_AS_OPTIONS:
                            expanded_po[YARRRML_OBJECTS][0][YARRRML_GATHER] = third_value.split('~')[0]
                            expanded_po[YARRRML_OBJECTS][0][YARRRML_GATHER_AS] = third_value.split('~')[1]
                    else:
                        expanded_po[YARRRML_OBJECTS][0][YARRRML_DATATYPE] = third_value

                expanded_predicateobjects.append(expanded_po)

            elif len(po) == 2:
                if isinstance(po[0], str):
                    po[0] = [po[0]]
                if isinstance(po[1], str):
                    po[1] = [po[1]]

                predicates_list, objects_list = po[0], po[1]

                for pred in predicates_list:
                    expanded_po = dict()
                    expanded_po[YARRRML_PREDICATES] = [{YARRRML_VALUE: pred}]
                    expanded_po[YARRRML_OBJECTS] = []

                    for obj in objects_list:
                        object_expansion = {}
                        if isinstance(obj, str) and '~' in obj:
                            obj_value, obj_type = obj.split('~')

                            if obj_type in YARRRML_GATHER_AS_OPTIONS:
                                object_expansion[YARRRML_GATHER_AS] = obj_type
                                object_expansion[YARRRML_GATHER] = obj_value
                            elif obj_type == "lang":
                                object_expansion[YARRRML_LANGUAGE] = obj_type
                                object_expansion[YARRRML_VALUE] = obj_value
                            else:
                                object_expansion[YARRRML_TYPE] = obj_type
                                object_expansion[YARRRML_VALUE] = obj_value
                        elif isinstance(obj, dict) and YARRRML_FUNCTION in obj:
                            object_expansion[YARRRML_FUNCTION] = obj[YARRRML_FUNCTION]
                            if YARRRML_PARAMETERS in obj:
                                object_expansion[YARRRML_PARAMETERS] = expand_parameters(obj[YARRRML_PARAMETERS])
                        else:
                            if isinstance(obj, dict):
                                if YARRRML_VALUE in obj:
                                    object_expansion.update(obj)
                            else:
                                object_expansion[YARRRML_VALUE] = obj

                        expanded_po[YARRRML_OBJECTS].append(object_expansion)

                    expanded_predicateobjects.append(expanded_po)
        elif isinstance(po, dict):
            expanded_po = {}

            if YARRRML_PREDICATES in po and YARRRML_OBJECTS in po:
                if isinstance(po[YARRRML_PREDICATES], str):
                    po[YARRRML_PREDICATES] = [po[YARRRML_PREDICATES]]
                if isinstance(po[YARRRML_OBJECTS], str):
                    po[YARRRML_OBJECTS] = [po[YARRRML_OBJECTS]]

                for pred in po[YARRRML_PREDICATES]:
                    expanded_po = dict()
                    if isinstance(pred, dict) and YARRRML_VALUE in pred:
                        expanded_po[YARRRML_PREDICATES] = po[YARRRML_PREDICATES]
                    else:
                        expanded_po[YARRRML_PREDICATES] = [{YARRRML_VALUE: pred}]

                    expanded_po[YARRRML_OBJECTS] = []
                    for obj in po[YARRRML_OBJECTS]:
                        object_expansion = {}
                        if isinstance(obj, dict):
                            if YARRRML_FUNCTION in obj and YARRRML_PARAMETERS in obj:
                                object_expansion[YARRRML_FUNCTION] = obj[YARRRML_FUNCTION]
                                object_expansion[YARRRML_PARAMETERS] = expand_parameters(obj[YARRRML_PARAMETERS])
                            elif YARRRML_MAPPING in obj or YARRRML_CONDITION in obj:
                                if YARRRML_MAPPING in obj:
                                    object_expansion[YARRRML_MAPPING] = obj[YARRRML_MAPPING]
                                if YARRRML_CONDITION in obj:
                                    condition_temp = obj[YARRRML_CONDITION]
                                    if isinstance(condition_temp, dict):
                                        condition_temp = [condition_temp]
                                    object_expansion[YARRRML_CONDITION] = [
                                        {
                                            **condition,
                                            YARRRML_PARAMETERS: expand_parameters(condition[YARRRML_PARAMETERS])
                                        } if YARRRML_PARAMETERS in condition else condition
                                        for condition in condition_temp
                                    ]
                            else:
                                if YARRRML_VALUE in obj:
                                    object_expansion[YARRRML_VALUE] = obj[YARRRML_VALUE]
                                if YARRRML_DATATYPE in obj:
                                    object_expansion[YARRRML_DATATYPE] = obj[YARRRML_DATATYPE]
                        elif isinstance(obj, list) and len(obj) == 2 and '~' in obj[1]:
                            object_expansion[YARRRML_VALUE] = obj[0]
                            object_expansion[YARRRML_LANGUAGE] = obj[1].split('~')[0]
                        elif isinstance(obj, list) and len(obj) == 2:
                            object_expansion[YARRRML_VALUE] = obj[0]
                            object_expansion[YARRRML_DATATYPE] = obj[1]
                        else:
                            object_expansion[YARRRML_VALUE] = obj

                        expanded_po[YARRRML_OBJECTS].append(object_expansion)

                    expanded_predicateobjects.append(expanded_po)

            elif 'p' in po and 'o' in po:
                if not isinstance(po['p'], list):
                    po['p'] = [po['p']]

                objects = po['o'] if isinstance(po['o'], list) else [po['o']]
                expanded_po[YARRRML_PREDICATES] = []

                for p in po['p']:
                    if YARRRML_FUNCTION in p:
                        expanded_predicate = {YARRRML_FUNCTION: p[YARRRML_FUNCTION]}
                        if YARRRML_PARAMETERS in p:
                            expanded_parameters = expand_parameters(p[YARRRML_PARAMETERS])
                            expanded_predicate.append(expanded_parameters)
                    else:
                        expanded_predicate = {YARRRML_VALUE: p}

                    expanded_po[YARRRML_PREDICATES].append(expanded_predicate)

                expanded_po[YARRRML_OBJECTS] = []

                for o in objects:
                    object_expansion = {}

                    if isinstance(o, dict):
                        if YARRRML_FUNCTION in o and YARRRML_PARAMETERS in o:
                            object_expansion[YARRRML_FUNCTION] = o[YARRRML_FUNCTION]
                            object_expansion[YARRRML_PARAMETERS] = expand_parameters(o[YARRRML_PARAMETERS])
                        elif YARRRML_MAPPING in o or YARRRML_CONDITION in o:
                            if YARRRML_MAPPING in o:
                                object_expansion[YARRRML_MAPPING] = o[YARRRML_MAPPING]
                            if YARRRML_CONDITION in o:
                                condition_temp = o[YARRRML_CONDITION]
                                if isinstance(condition_temp, dict):
                                    condition_temp = [condition_temp]
                                object_expansion[YARRRML_CONDITION] = [
                                    {
                                        **condition,
                                        YARRRML_PARAMETERS: expand_parameters(condition[YARRRML_PARAMETERS])
                                    } if YARRRML_PARAMETERS in condition else condition
                                    for condition in condition_temp
                                ]
                        elif YARRRML_GATHER in o:
                            if isinstance(o[YARRRML_GATHER], list):
                                object_expansion[YARRRML_GATHER] = []
                                for gather_value in o[YARRRML_GATHER]:
                                    if isinstance(gather_value, str) and '~' in gather_value:
                                        gather_entry = {
                                            YARRRML_GATHER: gather_value.split('~')[0],
                                            YARRRML_GATHER_AS: gather_value.split('~')[1]
                                        }
                                        object_expansion[YARRRML_GATHER].append(gather_entry)
                                    elif isinstance(gather_value, str):
                                        gather_entry = {
                                            YARRRML_GATHER: gather_value
                                        }
                                        object_expansion[YARRRML_GATHER].append(gather_entry)
                                    elif isinstance(gather_value, dict):
                                        gather_entry = {}
                                        if YARRRML_MAPPING in gather_value or YARRRML_CONDITION in gather_value:
                                            if YARRRML_MAPPING in gather_value:
                                                gather_entry[YARRRML_MAPPING] = gather_value[YARRRML_MAPPING]
                                            if YARRRML_CONDITION in gather_value:
                                                condition_temp = gather_value[YARRRML_CONDITION]
                                                if isinstance(condition_temp, dict):
                                                    condition_temp = [condition_temp]
                                                gather_entry[YARRRML_CONDITION] = [
                                                    {
                                                        **condition,
                                                        YARRRML_PARAMETERS: expand_parameters(condition[YARRRML_PARAMETERS])
                                                    } if YARRRML_PARAMETERS in condition else condition
                                                    for condition in condition_temp
                                                ]

                                            object_expansion[YARRRML_GATHER].append(gather_entry)
                                        else:
                                            object_expansion[YARRRML_GATHER].append(gather_value)

                            else:
                                object_expansion[YARRRML_GATHER] = o[YARRRML_GATHER]
                            object_expansion[YARRRML_GATHER_AS] = o[YARRRML_GATHER_AS]
                            if YARRRML_VALUE in o:
                                object_expansion[YARRRML_VALUE] = o[YARRRML_VALUE]
                            if YARRRML_TYPE in o:
                                object_expansion[YARRRML_TYPE] = o[YARRRML_TYPE]
                            if YARRRML_EMPTY in o:
                                object_expansion[YARRRML_EMPTY] = o[YARRRML_EMPTY]
                            if YARRRML_STRATEGY in o:
                                object_expansion[YARRRML_STRATEGY] = o[YARRRML_STRATEGY]
                        else:
                            object_expansion.update(o)
                    else:
                        object_expansion[YARRRML_VALUE] = o

                    if YARRRML_EMPTY in object_expansion:
                        if object_expansion[YARRRML_EMPTY] is True:
                            object_expansion[YARRRML_EMPTY] = "\"true\""
                        elif object_expansion[YARRRML_EMPTY] is False:
                            object_expansion[YARRRML_EMPTY] = "\"false\""
                    expanded_po[YARRRML_OBJECTS].append(object_expansion)

                expanded_predicateobjects.append(expanded_po)

            if 'graph' in po:
                expanded_po[YARRRML_GRAPHS] = [po['graph']]
            if YARRRML_GRAPHS in po:
                expanded_po[YARRRML_GRAPHS] = [po[YARRRML_GRAPHS]]

    return expanded_predicateobjects


def expand_parameters(parameters):
    expanded_parameters = list()
    for param in parameters:
        expanded_param = dict()
        if isinstance(param, list):
            if len(param) == 2:
                expanded_param[YARRRML_PARAMETER] = param[0]
                expanded_param[YARRRML_VALUE] = param[1]
            elif len(param) == 3:
                expanded_param[YARRRML_PARAMETER] = param[0]
                expanded_param[YARRRML_VALUE] = param[1]

        else:
            expanded_param = normalize_yaml(param)
        expanded_parameters.append(expanded_param)
    return expanded_parameters


def switch_mappings(data, external_sources, external_targets):
    sources_root = data.get(YARRRML_SOURCES, {})
    for source_name, source_value in sources_root.items():
        if source_name not in external_sources:
            external_sources[source_name] = copy.deepcopy(source_value)
        if YARRRML_MAPPINGS not in external_sources[source_name]:
            external_sources[source_name][YARRRML_MAPPINGS] = []
    targets_root = data.get(YARRRML_TARGETS, {})
    for target_name, target_value in targets_root.items():
        if target_name not in external_targets:
            external_targets[target_name] = copy.deepcopy(target_value)

    def replace_references(mapping_name, mapping_content):
        if YARRRML_SOURCES in mapping_content:
            expanded_sources = list()
            for source_ref in mapping_content[YARRRML_SOURCES]:
                if isinstance(source_ref, str) and source_ref in sources_root:
                    source = sources_root[source_ref]
                    expanded_sources.append(dict(source))
                    if source_ref in external_sources:
                        if mapping_name not in external_sources[source_ref][YARRRML_MAPPINGS]:
                            external_sources[source_ref][YARRRML_MAPPINGS].append(mapping_name)
                else:
                    expanded_sources.append(source_ref)


            mapping_content[YARRRML_SOURCES] = expanded_sources

        if YARRRML_SUBJECTS in mapping_content:
            for subject in mapping_content[YARRRML_SUBJECTS]:
                if isinstance(subject, dict):
                    if YARRRML_TARGETS in subject:
                        subject[YARRRML_TARGETS] = expand_targets_with_identifiers(subject[YARRRML_TARGETS], targets_root)

        if YARRRML_GRAPHS in mapping_content:
            for graph in mapping_content[YARRRML_GRAPHS]:
                if isinstance(graph, dict):
                    if YARRRML_TARGETS in graph:
                        graph[YARRRML_TARGETS] = expand_targets_with_identifiers(graph[YARRRML_TARGETS], targets_root)

        if YARRRML_PREDICATEOBJECT in mapping_content:
            po = mapping_content[YARRRML_PREDICATEOBJECT][0]
            if YARRRML_OBJECTS in po:
                for object in po[YARRRML_OBJECTS]:
                    if isinstance(object, dict):
                        if YARRRML_TARGETS in object:
                            object[YARRRML_TARGETS] = expand_targets_with_identifiers(object[YARRRML_TARGETS], targets_root)

    if YARRRML_MAPPINGS in data:
        for mapping_name, mapping_content in data[YARRRML_MAPPINGS].items():
            replace_references(mapping_name, mapping_content)

    if YARRRML_SOURCES in data:
        del data[YARRRML_SOURCES]
    if YARRRML_TARGETS in data:
        del data[YARRRML_TARGETS]

    return data


def expand_targets_with_identifiers(targets, root_targets):
    expanded_targets = list()
    if isinstance(targets, dict) or isinstance(targets,str):
        targets = [targets]
    for target in targets:
        if isinstance(target, str) and target in root_targets:
            expanded_targets.append(dict({target: root_targets[target]}))
        elif isinstance(target, list) and len(target) >= 1:
            expanded_targets.append(expand_targets(target))
        elif isinstance(target, dict):
            expanded_targets.append(target)
    return expanded_targets


def normalize(data, external_sources, external_targets):
    data = normalize_yaml(data)

    if data.get(YARRRML_MAPPINGS):
        for mapping in data.get(YARRRML_MAPPINGS):

            mapping_data = data.get(YARRRML_MAPPINGS).get(mapping)
            if YARRRML_PREDICATEOBJECT in mapping_data:
                for predicate_object_map in mapping_data.get(YARRRML_PREDICATEOBJECT):
                    if YARRRML_OBJECTS in predicate_object_map:
                        pass
                    else:
                        logger.error(
                            "There isn't a valid object key (object, objects, o) correctly specify in PON " + predicate_object_map)
                        raise Exception("Add or change the key of the object in the indicated POM")

                    if YARRRML_PREDICATES in predicate_object_map:
                        pass
                    else:
                        logger.error(
                            "There isn't a valid predicate key (predicate, predicates, p) correctly specify in PON " + predicate_object_map)
                        raise Exception("Add or change the key of the predicate in the indicated POM")

    switch_mappings(data, external_sources, external_targets)
    return data



from .constants import *
from .termmap import generate_rml_termmap


local_id = 0
def add_functions(yarrrml_data):
    global local_id
    functions = []
    for mapping in yarrrml_data.get(YARRRML_MAPPINGS):
        local_id = 0
        mapping_data = yarrrml_data.get(YARRRML_MAPPINGS).get(mapping)
        add_internal_function(mapping, mapping_data, functions)
    return functions

def add_internal_function(mapping_id, mapping_data, functions):
    global local_id
    keys = mapping_data.keys()
    for key in keys:
        if type(mapping_data[key]) is list:
            for i in range(len(mapping_data[key])):  # when the function is defined in POMs
                value = mapping_data[key][i]
                if YARRRML_FUNCTION in value:
                    function_return = transform_function(mapping_id, value, functions)
                    if type(function_return) is dict:
                        mapping_data[key][i] = function_return
                    elif type(function_return) is str:
                        mapping_data[key][i][YARRRML_FUNCTION] = function_return
                else:
                    if type(value) is list:
                        for v in value:
                            if type(v) is dict:
                                add_internal_function(mapping_id, v, functions)
                    elif type(value) is dict:
                        add_internal_function(mapping_id, value, functions)
        elif type(mapping_data[key]) is dict: # when the function is defined in the subject
            if YARRRML_FUNCTION in mapping_data[key]:
                function_return = transform_function(mapping_id,mapping_data[key], functions)
                if type(function_return) is dict:
                    mapping_data[key] = function_return
                elif type(function_return) is str:
                    mapping_data[key][YARRRML_FUNCTION] = function_return

def transform_function(mapping_id, function_value, functions):
    global local_id
    function_return = None
    if function_value[YARRRML_FUNCTION].replace(" ","").startswith(YARRRML_JOIN+"("):
        function_return = generate_extended_join(function_value[YARRRML_FUNCTION])
    elif function_value[YARRRML_FUNCTION] != YARRRML_EQUAL:
        function_id = "function_" + mapping_id
        function_return = function_id + "_" + str(local_id)
        functions.append(generate_function(function_value, function_id))
        local_id += 1  # different functions in the same TM

    return function_return

def generate_function(function_yarrrml_data, id_function):
    global local_id
    function = function_yarrrml_data[YARRRML_FUNCTION]
    if YARRRML_PARAMETERS in function_yarrrml_data:
        rml_function = ["<" + id_function +"_"+str(local_id)+"> a " + RML_EXECUTION_CLASS + ";\n\t"+RML_FUNCTION+" "+function+" ; \n"]
    else:
        function_yarrrml_data, function_name = split_in_line_function(function)
        rml_function = ["<" + id_function + "_" + str(
            local_id) + "> a " + RML_EXECUTION_CLASS + ";\n\t" + RML_FUNCTION + " " + function_name + " ; \n"]

    new_function = None
    if YARRRML_PARAMETERS in function_yarrrml_data:
        rml_function.append("\t"+RML_INPUT+"\n")
        parameters = function_yarrrml_data[YARRRML_PARAMETERS]
        for param in parameters:
            rml_function.append("\t\t[\n\t\t\ta "+RML_INPUT_CLASS + ";\n")
            if type(param) is list:
                param_extended = {YARRRML_PARAMETER: param[0], YARRRML_VALUE: param[1]}
            else:
                param_extended = param

            rml_function.append("\t\t\t"+RML_PARAMETER+" "+param_extended[YARRRML_PARAMETER] + ";\n")

            if YARRRML_VALUE in param_extended:
                if YARRRML_FUNCTION in param_extended[YARRRML_VALUE]:
                    local_id += 1  # new function definition within another functon
                    rml_function.append("\t\t\t" + RML_VALUE_MAP + "[\n\t\t\t\t" + RML_EXECUTION + " <"+id_function+"_"+str(local_id) +">;\n\t\t\t];\n")
                    new_function = generate_function(param_extended[YARRRML_VALUE],id_function)
                else:
                    rml_function.append(generate_rml_termmap(RML_VALUE_MAP, RML_VALUE_MAP_CLASS, param_extended[YARRRML_VALUE], "\t\t\t\t"))

            rml_function.append("\t\t],\n")


    final_function = "".join(rml_function)[0:-2]+".\n\n"
    if new_function:
        final_function += new_function

    return final_function



def split_in_line_function(function_in_line):
    parameters = {'parameters':[]}
    function_name = function_in_line.split("(")[0]
    function_in_line=function_in_line.replace(function_name, "").replace("(","",1).rsplit(")",1)[0]
    in_line_params = function_in_line.split(",")

    for param in in_line_params:
        extended_param = {}
        param_values = param.split("=",1)
        extended_param[YARRRML_PARAMETER] = param_values[0].strip()
        param_value = param_values[1].strip()

        if param_value.startswith("$("):
            extended_param[YARRRML_VALUE] = param_value
        else:
            value, internal_function_name = split_in_line_function(param_value)
            extended_param[YARRRML_VALUE] = {YARRRML_FUNCTION: internal_function_name, YARRRML_PARAMETERS: value[YARRRML_PARAMETERS]}

        parameters[YARRRML_PARAMETERS].append(extended_param)


    return parameters,function_name


def generate_extended_join(yarrrml_data):
    extended_join = {}
    data=yarrrml_data.replace(yarrrml_data.split("(")[0], "").replace("(","",1).rsplit(")",1)[0].split(",",1)
    if YARRRML_NON_ASSERTED in data[0]:
        extended_join[YARRRML_NON_ASSERTED] = data[0].split("=")[1]
    elif YARRRML_QUOTED in data[0]:
        extended_join[YARRRML_QUOTED] = data[0].split("=")[1]
    elif YARRRML_MAPPING in data[0] and "=" in data[0]:
        extended_join[YARRRML_MAPPING] = data[0].split("=")[1]
    else:
        extended_join[YARRRML_MAPPING] = data[0]

    equals = data[1].split(YARRRML_EQUAL)
    equals = list(filter(lambda val: val != ' ', equals))

    if equals:
        extended_join[YARRRML_CONDITION] = []

    for value in equals:
        parameters = []
        for param_name, condition in zip(['str1', 'str2'], value.replace("(", "", 1).rsplit(")", 1)[0].split(",")):
            parameters.append({
                YARRRML_PARAMETER: param_name,
                YARRRML_VALUE: condition
            })
        extended_join[YARRRML_CONDITION].append({
            YARRRML_FUNCTION: YARRRML_EQUAL,
            YARRRML_PARAMETERS: parameters
        })

    return extended_join






import os
import re
import rdflib
from .constants import *
from ruamel.yaml import YAML




def add_source(data, mapping, external_sources={}):
    source_template = "\t" + RML_LOGICAL_SOURCE + " [\n\t\ta " + RML_LOGICAL_SOURCE_CLASS + \
                      ";\n\t\t" + RML_SOURCE + " "
    final_list = []
    sources = data.get(YARRRML_MAPPINGS).get(mapping).get(YARRRML_SOURCES)
    external_references_list = []
    for source in sources:
        db_identifier = mapping
        for external_Source in external_sources:
            external_source = {k: v for k, v in external_sources[external_Source].items() if k != 'mappings'}
            if source == external_source:
                db_identifier = external_Source
        if YARRRML_ACCESS in source:
            if YARRRML_QUERY in source:
                final_list.append(source_template + database_source(mapping, source, db_identifier))
            elif YARRRML_STRUCTURE_DEFINER in source:
                source, external_references = add_in_memory_source(mapping,source)
                final_list.append(source_template + source)
                if external_references is not None:
                    external_references_list.append(external_references)
            else:
                final_list.append(source_template + add_source_full(mapping, source))
        elif type(source) is list: # ToDo: review for in_memory sources
            if "$(" in source[0]:
                source, external_references = add_in_memory_source(mapping, source)
                final_list.append(source_template + source)
                if external_references is not None:
                    external_references_list.append(external_references)
            else:
                final_list.append(source_template + add_source_simplified(mapping, source)) # this probably needs to be removed
        else:
            raise Exception("ERROR: source " + source + " in mapping " + mapping + " not valid")
    return final_list, external_references_list


def add_table(data, mapping):
    table_template = "\t" + R2RML_LOGICAL_TABLE + " [\n\t\ta " + R2RML_LOGICAL_TABLE_CLASS + \
                     ";\n\t\t"

    final_list = []
    sources = data.get(YARRRML_MAPPINGS).get(mapping).get(YARRRML_SOURCES)
    for source in sources:
        sql_version = False
        db_identifier = mapping
        if YARRRML_ACCESS in source and YARRRML_QUERY in source:
            r2rml_access = database_source(mapping, source, db_identifier)
            sql_version = True
        elif YARRRML_QUERY in source:
            r2rml_access = R2RML_SQL_QUERY + " \"" + source.get(YARRRML_QUERY).replace("\n", " ").replace("\"",
                                                                                                          "\\\"") + "\""
        elif YARRRML_TABLE in source:
            r2rml_access = R2RML_TABLE_NAME + " \"" + source.get(YARRRML_TABLE) + "\""
        else:
            raise Exception("ERROR: table or query is not provided in " + source + " of mapping " + mapping)
        if not sql_version:
            if YARRRML_QUERY_FORMULATION in source:
                r2rml_access += ";\n\t\t" + R2RML_SQL_VERSION + " rr:" + source.get(YARRRML_QUERY_FORMULATION).upper()
            else:
                r2rml_access += ";\n\t\t" + R2RML_SQL_VERSION + " rr:SQL2008"
            r2rml_access += "\n\t];\n"
        final_list.append(table_template + r2rml_access)
    return final_list



def add_source_simplified(mapping, source):
    source_rdf = ""
    file_path = re.sub("~.*", "", source[0])
    reference_formulation = source[0].split('~')[1]
    source_extension = os.path.splitext(file_path)[1].replace(".","")
    ref_formulation_rml = YARRRML_REFERENCE_FORMULATIONS[reference_formulation]

    if switch_in_reference_formulation(reference_formulation, source_extension) != source_extension:
        raise Exception(
            "ERROR: mismatch extension and referenceFormulation in source " + source + " in mapping " + mapping)
    else:
        if len(source) == 1:  # do not have iterator
            if source_extension == "csv" or source_extension == "SQL2008" or source_extension == "xlsx":
                source_rdf += '"' + file_path + '"' + ";\n" + "\t\t" + RML_REFERENCE_FORMULATION + " ql:" \
                              + ref_formulation_rml + "\n" + "\t];\n"
            else:
                raise Exception("ERROR: source " + source + " in mapping " + mapping + " has no iterator")
        else:  # source[1] is the iterator for json and xml
            source_rdf += "\"" + file_path + "\";\n\t\t" + RML_REFERENCE_FORMULATION + " ql:" \
                          + ref_formulation_rml + ";\n\t\t" + RML_ITERATOR + " \"" \
                          + source[1] + "\";\n\t];\n"
    return source_rdf

def add_in_memory_source(mapping, source):
    external_reference_formulation = None
    if type(source) is list:
        source = extend_in_memory(source)
    source_rdf = "[\n\t\t\ta " + SD_DATASET_SPEC + ";\n\t\t\t"

    access = str(source.get(YARRRML_ACCESS)).replace("$(","").replace(")","")
    source_rdf += SD_NAME + " \"" + access + "\";\n"

    if YARRRML_SOFTWARE_SPECIFICATION in source:
        source_rdf += "\t\t\t" + SD_HAS_DATA_TRANSFORMATION + "[\n\t\t\t\t"

        if YARRRML_SOFTWARE_REQUIREMENTS in source.get(YARRRML_SOFTWARE_SPECIFICATION):
            software_requirements = str(source.get(YARRRML_SOFTWARE_SPECIFICATION)[YARRRML_SOFTWARE_REQUIREMENTS])
            source_rdf += SD_HAS_SOFTWARE_REQUIREMENTS + " \""+ software_requirements +"\";\n\t\t\t\t"

        if YARRRML_PROGRAMMING_LANGUAGE in source.get(YARRRML_SOFTWARE_SPECIFICATION):
            programming_language = str(source.get(YARRRML_SOFTWARE_SPECIFICATION)[YARRRML_PROGRAMMING_LANGUAGE])
            source_rdf += SD_HAS_SOURCE_CODE +"[\n\t\t\t\t\t" + SD_PROGRAMMING_LANGUAGE + " \"" + programming_language + "\";"
            source_rdf +="\n\t\t\t\t];\n"

        source_rdf += "\t\t\t];\n"
    source_rdf += "\t\t];\n"

    if YARRRML_ITERATOR in source:
        source_rdf += "\t\t" + RML_ITERATOR + " \"" + source.get(YARRRML_ITERATOR) + "\";\n"

    if YARRRML_REFERENCE_FORMULATION in source:
        reference_formulation = str(source.get(YARRRML_REFERENCE_FORMULATION))
        source_rdf += "\t\t" + RML_REFERENCE_FORMULATION + " ql:"+ reference_formulation + ";\n"
        external_reference_formulation = "ql:" + reference_formulation + " a " + RML_REFERENCE_FORMULATION_CLASS +";\n"
        external_reference_formulation +="\t" + KG4DI_DEFINED_BY +" \""+source.get(YARRRML_STRUCTURE_DEFINER) +"\"."
    source_rdf += "\t];\n"



    return source_rdf, external_reference_formulation

def extend_in_memory(source):
    features = source[0].split("~")
    access = features[0]
    defined_by = features[1].split("-")[0]
    reference_formulation = features[1].split("-")[1]
    extended_source = {"access": access, "structureDefiner": defined_by, "referenceFormulation": reference_formulation}
    if len(source) == 2:
        extended_source["iterator"] = source[1]
    return extended_source


def add_source_full(mapping, source):
    source_rdf = ""

    access = str(source.get(YARRRML_ACCESS))
    extension = os.path.splitext(access)[1][1:]
    if YARRRML_REFERENCE_FORMULATION in source:
        reference_formulation = str(source.get(YARRRML_REFERENCE_FORMULATION))
        format_from_reference = switch_in_reference_formulation(reference_formulation.lower())
        ref_formulation_rml = YARRRML_REFERENCE_FORMULATIONS[reference_formulation]

        if extension != format_from_reference or format_from_reference == "ERROR":
            raise Exception("ERROR: not referenceFormulation found or mismatch between the format and "
                            "referenceFormulation in source " + access + "in mapping " + mapping)
        if YARRRML_ITERATOR in source:
            source_iterator = str(source.get(YARRRML_ITERATOR))

            source_rdf += "\"" + access + "\";\n\t\t" + RML_REFERENCE_FORMULATION + " ql:" \
                          + ref_formulation_rml + ";\n\t\t" + RML_ITERATOR + " \"" \
                          + source_iterator + "\";\n\t];\n"
        else:
            if extension == "csv" or extension == "SQL2008":
                source_rdf += "\"" + access + "\";\n\t\t" + RML_REFERENCE_FORMULATION + " ql:" \
                              + ref_formulation_rml + "\n\n\t];\n"
            else:
                raise Exception("ERROR: source " + access + "in mapping " + mapping + " has no referenceFormulation")

    else:
        if extension == "csv":
            source_rdf += "\"" + access + "\";\n\n\t];\n"
        else:
            raise Exception("ERROR: source " + access + "in mapping " + mapping + " has no referenceFormulation")

    return source_rdf


def database_source(mapping, source, db_identifier):
    source_rdf = ""
    if YARRRML_ACCESS in source:
        if YARRRML_CREDENTIALS in source:
            if YARRRML_TYPE in source:
                source_rdf += "<DataSource_" + str(db_identifier) + ">;\n\t\t"
                if YARRRML_QUERY in source:
                    source_rdf += RML_QUERY + " \"" + source.get(YARRRML_QUERY).replace("\n", " ").replace("\"",
                                                                                                           "\\\"") + "\""
                elif YARRRML_TABLE in source:
                    source_rdf += R2RML_TABLE_NAME + " \"" + source.get(YARRRML_TABLE) + "\""
                if YARRRML_REFERENCE_FORMULATION in source:
                    source_rdf += ";\n\t\t" + RML_REFERENCE_FORMULATION + " ql:" \
                                  + switch_in_reference_formulation(source.get(YARRRML_REFERENCE_FORMULATION)).upper()
                if YARRRML_QUERY_FORMULATION in source:
                    source_rdf += ";\n\t\t" + R2RML_SQL_VERSION + " rr:" + source.get(YARRRML_QUERY_FORMULATION).upper()
                else:
                    source_rdf += ";\n\t\t" + R2RML_SQL_VERSION + " rr:SQL2008"
                source_rdf += "\n\t];\n"
        else:
            raise Exception("ERROR: no credentials to get access to source in mapping " + mapping)
    else:
        raise Exception("ERROR: no access to the source in mapping " + mapping)

    return source_rdf


def switch_in_reference_formulation(value, source_extension=None):
    value = value.lower()
    if "json" in value:
        if "path" in value:
            switcher = "json"
        else:
            switcher = "jsonpath"
    elif "x" in value:
        if "path" in value:
            switcher = "xml"
        else:
            switcher = "xpath"
    elif source_extension:
        if source_extension == "xlsx":
            switcher = "xlsx"
        else:
            switcher = value
    else:
        switcher = value
    return switcher


def generate_database_connections(data, external_sources):
    database = []
    for mapping in data.get(YARRRML_MAPPINGS):
        sources = data.get(YARRRML_MAPPINGS).get(mapping).get(YARRRML_SOURCES)
        for source in sources:
            db_identifier = mapping
            for external_Source in external_sources:
                external_source = {k: v for k, v in external_sources[external_Source].items() if k != 'mappings'}
                if source == external_source:
                    db_identifier = external_Source
                    break
            if YARRRML_QUERY in source and YARRRML_ACCESS in source:
                db_type = source.get(YARRRML_TYPE)
                if db_type in YARRRML_DATABASES_DRIVER:
                    driver = YARRRML_DATABASES_DRIVER[db_type]
                else:
                    driver = None
                access = source.get(YARRRML_ACCESS)
                username = source.get(YARRRML_CREDENTIALS).get(YARRRML_USERNAME)
                password = source.get(YARRRML_CREDENTIALS).get(YARRRML_PASSWORD)

                if driver is None:
                    connection_string = "<DataSource_" + str(db_identifier) + "> a " + D2RQ_DATABASE_CLASS + ";\n\t" \
                                        + D2RQ_DSN + " \"" + access + "\";\n\t" \
                                        + D2RQ_USER + " \"" + username + "\";\n\t" \
                                        + D2RQ_PASS + " \"" + password + "\".\n\n"
                else:
                    connection_string = "<DataSource_" + str(db_identifier) + "> a " + D2RQ_DATABASE_CLASS + ";\n\t" \
                                        + D2RQ_DSN + " \"" + access + "\";\n\t" \
                                        + D2RQ_DRIVER + " \"" + driver + "\";\n\t" \
                                        + D2RQ_USER + " \"" + username + "\";\n\t" \
                                        + D2RQ_PASS + " \"" + password + "\".\n\n"

                if connection_string not in database:
                    database.append(connection_string)

    return database


def add_inverse_source(tm, rdf_mapping, mapping_format):
    try:
        query = f'SELECT ?source  WHERE {{ <{tm}> {R2RML_LOGICAL_TABLE}|{RML_LOGICAL_SOURCE} ?source . }} '
        source = [tm[rdflib.Variable('source')] for tm in rdf_mapping.query(query).bindings][0]
    except Exception as e:
        logger.error("Logical Source or Logical Table is not defined in the mapping")
        logger.error(str(e))

    if mapping_format == R2RML_URI:
        yarrrml_source = get_logical_table(source, rdf_mapping)
    else:
        yarrrml_source = get_logical_source(source, rdf_mapping)

    return yarrrml_source


def get_logical_table(logical_table_id, rdf_mapping):
    table_name = rdf_mapping.value(subject=logical_table_id, predicate=rdflib.Namespace(R2RML_URI).tableName)
    sql_query = rdf_mapping.value(subject=logical_table_id, predicate=rdflib.Namespace(R2RML_URI).sqlQuery)
    sql_version = rdf_mapping.value(subject=logical_table_id, predicate=rdflib.Namespace(R2RML_URI).sqlVersion)

    if table_name is None and sql_query is None:
        logger.error("Mapping does not define neither tableName nor sqlQuery")
        raise Exception()
    yarrrml_source = {}
    if table_name:
        yarrrml_source["table"] = table_name.value
    elif sql_query:
        yarrrml_source["query"] = sql_query.value

    if sql_version:
        yarrrml_source["queryFormulation"] = sql_version.toPython().replace(R2RML_URI, '').lower()

    return yarrrml_source


def get_logical_source(logical_source_id, rdf_mapping):
    yaml = YAML()
    source = rdf_mapping.value(subject=logical_source_id, predicate=rdflib.Namespace(RML_URI).source)
    iterator = rdf_mapping.value(subject=logical_source_id, predicate=rdflib.Namespace(RML_URI).iterator)
    reference_formulation = rdf_mapping.value(subject=logical_source_id,
                                              predicate=rdflib.Namespace(RML_URI).referenceFormulation)
    sql_query = rdf_mapping.value(subject=logical_source_id, predicate=rdflib.Namespace(R2RML_URI).sqlQuery)
    sql_version = rdf_mapping.value(subject=logical_source_id, predicate=rdflib.Namespace(R2RML_URI).sqlVersion)

    if source is None:
        logger.error("Mapping does not define source access")
        raise Exception()

    if source and reference_formulation and iterator:
        yarrrml_source = [source.value + '~' + reference_formulation.toPython().replace(QL_URI, '').lower(), iterator.value]
    elif source and sql_query:
        # this means a database source
        source_dict = {"query": sql_query.value, "source": source.value}
        if reference_formulation:
            source_dict["referenceFormulation"] = reference_formulation.toPython().replace(QL_URI, '').lower()
        if sql_version:
            source_dict["queryFormulation"] = sql_version.toPython().replace(R2RML_URI, '').lower()
        yarrrml_source = source_dict
    elif source and reference_formulation:
        yarrrml_source = [source.value + '~' + reference_formulation.toPython().replace(QL_URI, '').lower()]
    else:
        if source.endsWith(".csv"):
            yarrrml_source = [source.value + '~csv']
        else:
            yarrrml_source = [source.value]

    if type(yarrrml_source) is list:
        yarrrml_source = yaml.seq(yarrrml_source)
        yarrrml_source.fa.set_flow_style()

    return yarrrml_source



from .constants import *
from .termmap import generate_rml_termmap, generate_cc_termmap
from .graph import add_inverse_graph
import rdflib


## Adds a subject or set of subjects to the given TriplesMap
def add_subject(data, mapping, mapping_format):
    rml_subjects = []

    if YARRRML_SUBJECTS in data.get(YARRRML_MAPPINGS).get(mapping):
        subject = data.get(YARRRML_MAPPINGS).get(mapping).get(YARRRML_SUBJECTS)
    else:
        raise Exception("ERROR: no subjects in mapping " + mapping)

    for individual_subject in subject:
        if YARRRML_QUOTED in individual_subject or YARRRML_NON_ASSERTED in individual_subject:
            if YARRRML_CONDITION in individual_subject:
                from .predicateobject import ref_mapping
                if YARRRML_NON_ASSERTED in individual_subject:
                    subject_termmap = ref_mapping(data, mapping, individual_subject, YARRRML_NON_ASSERTED, STAR_QUOTED,
                                                  mapping_format)
                else:
                    subject_termmap = ref_mapping(data, mapping, individual_subject, YARRRML_QUOTED, STAR_QUOTED,
                                                  mapping_format)
                subject_termmap = subject_termmap.replace(R2RML_REFOBJECT_CLASS, STAR_CLASS).replace(STAR_OBJECT,
                                                                                                     STAR_SUBJECT).replace(
                    "\t\t\t\t", "\t\t\t\t\t").replace("\t\t", "\t")
            else:
                subject_termmap = generate_rml_termmap(STAR_SUBJECT, STAR_CLASS, individual_subject, "\t\t")
        elif mapping_format == STAR_URI:
            subject_termmap = generate_rml_termmap(STAR_SUBJECT, R2RML_SUBJECT_CLASS, individual_subject, "\t\t")
        elif YARRRML_GATHER in individual_subject:
            subject_termmap = generate_cc_termmap(STAR_SUBJECT, R2RML_SUBJECT_CLASS, individual_subject, "\t\t\t", mapping_format) + "\n\t\t];\n"

        else:
            subject_value = individual_subject
            if YARRRML_VALUE in individual_subject and type(individual_subject) is dict:
                subject_value = individual_subject.get(YARRRML_VALUE)
            elif YARRRML_FUNCTION in individual_subject:
                subject_value = individual_subject.get(YARRRML_FUNCTION)

            subject_termmap = generate_rml_termmap(R2RML_SUBJECT, R2RML_SUBJECT_CLASS, subject_value, "\t\t",
                                                   mapping_format)

            if YARRRML_FUNCTION in individual_subject:
                subject_termmap = subject_termmap.replace(R2RML_CONSTANT + " \"" + subject_value + "\"",
                                                          RML_EXECUTION + " <" + individual_subject.get(
                                                              YARRRML_FUNCTION) + ">")

            if YARRRML_TARGETS in individual_subject:
                subject_termmap = subject_termmap[0:-3] + "\t" + RML_LOGICAL_TARGET + " <" + individual_subject[
                    YARRRML_TARGETS] + ">\n\t];\n"

            if isinstance(individual_subject, dict) and YARRRML_TYPE in individual_subject:
                if individual_subject.get(YARRRML_TYPE) == YARRRML_BLANK:
                    subject_termmap = subject_termmap[
                                      0:-3] + "\t" + R2RML_TERMTYPE + " " + R2RML_BLANK_NODE + "\n\t];\n"

        rml_subjects.append(subject_termmap)

    if YARRRML_GRAPHS in data.get(YARRRML_MAPPINGS).get(mapping):
        graphs = data.get(YARRRML_MAPPINGS).get(mapping).get(YARRRML_GRAPHS)
    else:
        graphs = []

    if type(graphs) != list:
        graphs = [graphs]

    for graph in graphs:
        graph_value = graph
        if YARRRML_VALUE in graph:
            graph_value = graph[YARRRML_VALUE]
        graph_termmap = generate_rml_termmap(R2RML_GRAPH_MAP, R2RML_GRAPH_CLASS, graph_value, "\t\t\t")
        if YARRRML_TARGETS in graph:
            graph_termmap = graph_termmap[0:-3] + "\t" + RML_LOGICAL_TARGET + " <" + graph[
                YARRRML_TARGETS] + ">\n\t\t];\n"
        rml_subjects = list(map(lambda subject: subject[0:-4] + graph_termmap + "\t];\n", rml_subjects))

    return rml_subjects


def add_inverse_subject(tm, rdf_mapping):
    query = f'SELECT ?subject_value  WHERE {{ <{tm}> {R2RML_SUBJECT} ?subject . ' \
            f'?subject {R2RML_TEMPLATE}|{R2RML_COLUMN}|{RML_REFERENCE}|{R2RML_CONSTANT} ?subject_value' \
            f'}} '

    subject = [tm[rdflib.Variable('subject_value')].value for tm in rdf_mapping.query(query).bindings]
    if len(subject) > 1:
        logger.error("There are more than one subject in your mapping")
        raise Exception("A mapping defines one and only one subjectMap")
    else:
        subject = subject[0]
    # this means it's a column
    if not subject.startswith("http"):
        subject = '$(' + subject + ')'
    else:
        if '{' in subject:
            subject = subject.replace('{', "$(").replace('}', ')')

    yarrrml_subject = {'s': subject}

    query = f'SELECT ?class_value  WHERE {{ <{tm}> {R2RML_SUBJECT} ?subject . ' \
            f'?subject {R2RML_CLASS} ?class_value }} '
    classes = [tm[rdflib.Variable('class_value')] for tm in rdf_mapping.query(query).bindings]

    query = f'SELECT ?graph_value  WHERE {{ <{tm}> {R2RML_SUBJECT} ?subject . ' \
            f'{{ ?subject {R2RML_GRAPH} ?graph_value .  }}' \
            f' UNION ' \
            f' {{ ?subject {R2RML_GRAPH_MAP} ?graph_map .' \
            f' ?graph_map {R2RML_TEMPLATE}|{R2RML_COLUMN}|{RML_REFERENCE}|{R2RML_CONSTANT} ?graph_value  }} }} '
    graph_values = [tm[rdflib.Variable('graph_value')] for tm in rdf_mapping.query(query).bindings]

    if graph_values:
        yarrrml_subject.update(add_inverse_graph(graph_values))

    return yarrrml_subject, classes



from .constants import *
from .mapping import add_prefix, add_mapping, add_inverse_prefix, get_non_asserted_mappings, merge_mapping_section_by_key
from .source import add_source, generate_database_connections, add_table, add_inverse_source
from .subject import add_subject, add_inverse_subject
from .predicateobject import add_predicate_object_maps, add_inverse_pom
from .target import add_logical_targets
from .function import add_functions
from .normalization import normalize
import rdflib
import ruamel.yaml as yaml


def translate(yarrrml_data, mapping_format=RML_URI):
    logger.info("Translating YARRRML mapping to [R2]RML")
    external_sources = {}
    external_targets = {}
    yarrrml_data = normalize(yarrrml_data, external_sources, external_targets)
    rml_mapping = [add_prefix(yarrrml_data)]
    rml_mapping.extend(generate_database_connections(yarrrml_data, external_sources))
    rml_mapping.extend(add_logical_targets(yarrrml_data, external_targets))
    rml_mapping.extend(add_functions(yarrrml_data))
    external_references = []
    try:
        mappings, mapping_format = get_non_asserted_mappings(yarrrml_data, mapping_format)
        for mapping in yarrrml_data.get(YARRRML_MAPPINGS):
            if mapping_format == R2RML_URI:
                source_list = add_table(yarrrml_data, mapping)
            else:
                source_list, external_references = add_source(yarrrml_data, mapping, external_sources)
            subject_list = add_subject(yarrrml_data, mapping, mapping_format)
            pred = add_predicate_object_maps(yarrrml_data, mapping, mapping_format)
            it = 0
            for source in source_list:
                for subject in subject_list:
                    map_aux = add_mapping(mapping, mappings, it)
                    if type(source) is list:
                        rml_mapping.append(map_aux + source[0] + subject + pred + source[1])
                    else:
                        rml_mapping.append(map_aux + source + subject + pred)
                    rml_mapping[len(rml_mapping) - 1] = rml_mapping[len(rml_mapping) - 1][:-2]
                    rml_mapping.append(".\n\n\n")
                    it = it + 1

            external_refs = list(dict.fromkeys(external_references))
            for ref in external_refs:
                rml_mapping.append(ref)

        logger.info("RML content is created!")
        rml_mapping_string = "".join(rml_mapping)
        try:
            graph = rdflib.Graph()
            graph.parse(data=rml_mapping_string, format="turtle")
            logger.info("Mapping has been syntactically validated.")
        except Exception as e:
            logger.error("ERROR: There is a syntactic error in the generated mapping")
            logger.error(str(e))
            return None
    except Exception as e:
        logger.error("ERROR: The YARRRML mapping has not been translated")
        logger.error(str(e))
        return None

    logger.info("Translation has finished successfully.")

    return rml_mapping_string


def inverse_translation(rdf_mapping, mapping_format=RML_URI):
    logger.info("Translating [R2]RML mappings to YARRRML")
    yarrrml_mapping = {YARRRML_PREFIXES: {}, YARRRML_MAPPINGS: {}}
    rdf_mapping.bind('rml', rdflib.term.URIRef(RML_URI))
    rdf_mapping.bind('rr', rdflib.term.URIRef(R2RML_URI))
    rdf_mapping.bind('ql', rdflib.term.URIRef(QL_URI))
    yarrrml_mapping[YARRRML_PREFIXES] = add_inverse_prefix(rdf_mapping)
    query = f'SELECT ?triplesMap WHERE {{ ?triplesMap {RML_LOGICAL_SOURCE}|{R2RML_LOGICAL_TABLE} ?source . }} '
    triples_map = [tm[rdflib.Variable('triplesMap')] for tm in rdf_mapping.query(query).bindings]

    for tm in triples_map:
        tm_name = tm.split("/")[-1].split("#")[-1]
        yarrrml_tm = {YARRRML_SOURCES: [add_inverse_source(tm, rdf_mapping, mapping_format)]}
        subject, classes = add_inverse_subject(tm, rdf_mapping)
        yarrrml_tm.update(subject)
        yarrrml_tm['po'] = add_inverse_pom(tm, rdf_mapping, classes, yarrrml_mapping[YARRRML_PREFIXES])
        yarrrml_mapping[YARRRML_MAPPINGS][tm_name] = yarrrml_tm

    logger.info("Translation has finished successfully.")
    return yarrrml_mapping

def merge_mappings(yarrrrml_list):
    combined_mapping = {YARRRML_MAPPINGS:{}}
    prefixes_list = []
    for mapping in yarrrrml_list:
            prefixes_list.append(mapping[YARRRML_PREFIXES])
    combined_mapping = combined_mapping | merge_mapping_section_by_key(YARRRML_PREFIXES, prefixes_list)

    triples_map_ids = yarrrrml_list[0][YARRRML_MAPPINGS].keys()
    for individual_id in triples_map_ids:
        mapping_content_list = []
        for mapping in yarrrrml_list:
            if individual_id in mapping[YARRRML_MAPPINGS]:
                mapping_content_list.append(mapping[YARRRML_MAPPINGS][individual_id])
        combined_mapping[YARRRML_MAPPINGS] = combined_mapping[YARRRML_MAPPINGS] | merge_mapping_section_by_key(individual_id, mapping_content_list)

    string_content = yaml.dump(combined_mapping)
    return string_content



from .constants import *


## return the type of TermMap based on the input text
def get_termmap_type(text, mapping_format):
    if "$(" in text and ")" in text:
        if text[0] == "$" and text[len(text) - 1] == ")" and text.count("$(") == 1:
            if mapping_format == R2RML_URI:
                return R2RML_COLUMN
            else:
               return RML_REFERENCE
        else:
            return R2RML_TEMPLATE
    elif YARRRML_QUOTED in text or YARRRML_NON_ASSERTED in text:
        return STAR_QUOTED
    else:
        return R2RML_CONSTANT

def generate_cc_termmap_text(text, mapping_format=RML_URI):
    from .mapping import prefixes
    term_map = get_termmap_type(text[YARRRML_GATHER], mapping_format)
    text = text[YARRRML_GATHER]
    if term_map == R2RML_TEMPLATE:
        text = generate_rml_template(text)
        text = text.replace('"', r'\"')
        if ":" in text:
            text_prefix_split = text.split(":")
            if text_prefix_split[0] in prefixes:
                text = prefixes[text_prefix_split[0]] + text_prefix_split[1]
    elif term_map == RML_REFERENCE or term_map == R2RML_COLUMN:
        text = text.replace("$(", "").replace(")", "")
        text = text.replace('"', r'\"')
        text = f'"{text}"'
    elif term_map == R2RML_CONSTANT and text == "a":
        text = RDF_TYPE
    return term_map, text

def generate_rml_termmap_text(text, mapping_format=RML_URI):
    from .mapping import prefixes
    term_map = get_termmap_type(text, mapping_format)
    if term_map == R2RML_TEMPLATE:
        text = generate_rml_template(text)
        text = text.replace('"', r'\"')
        if ":" in text:
            text_prefix_split = text.split(":")
            if text_prefix_split[0] in prefixes:
                text = prefixes[text_prefix_split[0]] + text_prefix_split[1]
    elif term_map == RML_REFERENCE or term_map == R2RML_COLUMN:
        text = generate_rml_template(text).replace("{","").replace("}","")
        text = text.replace('"', r'\"')
    elif term_map == R2RML_CONSTANT and text == "a":
        text = RDF_TYPE
    return term_map, text

def generate_cc_termmap(rml_property, rml_class, content, indentation, mapping_format=RML_URI):
    template = indentation[0:-1] + rml_property + " [\n" + indentation + "a " + rml_class + ";\n"
    if isinstance(content, dict) and 'value' in content:
        value_text = content.get('value', '')
        term_map, value_text = generate_rml_termmap_text(value_text, mapping_format)
        if term_map == STAR_QUOTED:
            if 'quoted' in value_text:
                template +=indentation + term_map + " <" + value_text[YARRRML_QUOTED] + "_0>;\n"
            else:
                template +=indentation +  term_map + " <" + value_text[YARRRML_NON_ASSERTED] + "_0>;\n"
        elif term_map != "rr:constant":
            template +=indentation +  term_map + " \"" + value_text + "\";\n"
        else:
            if value_text.startswith("http"):
                template +=indentation +  term_map + " <" + value_text + ">;\n"
            else:
                if ":" in value_text or "<" in value_text:
                    template +=indentation +  term_map + " " + value_text + ";\n"
                else:
                    template +=indentation +  term_map + " \"" + value_text + "\";\n"
    if isinstance(content[YARRRML_GATHER], str):
        term_map, text = generate_cc_termmap_text(content, mapping_format)
        if YARRRML_EMPTY in content:
            if content[YARRRML_EMPTY] == "\"true\"":
                template += indentation + RML_CC_EMPTY +" " + "true" + ";\n"
            elif content[YARRRML_EMPTY] == "\"false\"":
                template += indentation + RML_CC_EMPTY +" " + "false" + ";\n"
        template += indentation + RML_CC_GATHER + " ( [ " + term_map + " " + text + "; ] ) ;\n"
        if YARRRML_GATHER_AS in content and content[YARRRML_GATHER_AS] in YARRRML_GATHER_AS_OPTIONS:
            template += indentation + RML_CC_GATHER_AS + " rdf:" + content[YARRRML_GATHER_AS].capitalize() + ";\n"
    elif isinstance(content[YARRRML_GATHER], list):
        template += indentation + RML_CC_GATHER + " (\n"
        for gather_item in content[YARRRML_GATHER]:
            term_map, text = generate_cc_termmap_text(gather_item, mapping_format)
            if YARRRML_GATHER_AS in gather_item:
                template += indentation + "\t" + "[\n" + indentation + "\t\t" + RML_CC_GATHER + " ( [ " + term_map + " " + text + "; ] ) ;\n"
                template += indentation + "\t\t" + RML_CC_GATHER_AS + " rdf:" + gather_item[YARRRML_GATHER_AS].capitalize() + ";\n"
                template += indentation + "\t]\n"
            else:
                template += indentation + " [ " + term_map + " " + text + "; ] \n"

        template += indentation + ") ;\n"

        if YARRRML_GATHER_AS in content and content[YARRRML_GATHER_AS] in YARRRML_GATHER_AS_OPTIONS:
            template += indentation + RML_CC_GATHER_AS + " rdf:" + content[YARRRML_GATHER_AS].capitalize() + ";\n"
        if YARRRML_STRATEGY in content and content[YARRRML_STRATEGY] in ['append','cartesianProduct']:
            template += indentation + RML_CC_STRATEGY + " rml:" + content[YARRRML_STRATEGY] + ";\n"
    return template


## Generates a TermMap (subject, predicate, object) based on the property, class and the text
def generate_rml_termmap(rml_property, rml_class, text, indentation, mapping_format=RML_URI):
    template = indentation[0:-1] + rml_property + " [\n" + indentation + "a " + rml_class + ";\n" + indentation
    if isinstance(text, dict) and 'value' in text:
        text = text.get('value', '')
    term_map, text = generate_rml_termmap_text(text, mapping_format)

    if term_map == STAR_QUOTED:
        if 'quoted' in text:
            template += term_map + " <" + text[YARRRML_QUOTED] + "_0>;\n" + indentation[0:-1] + "];\n"
        else:
            template += term_map + " <" + text[YARRRML_NON_ASSERTED] + "_0>;\n" + indentation[0:-1] + "];\n"
    elif term_map != "rr:constant":
        template += term_map + " \"" + text + "\";\n"+indentation[0:-1]+"];\n"
    else:
        if text.startswith("http"):
            template += term_map + " <" + text + ">;\n" + indentation[0:-1] + "];\n"
        else:
            if ":" in text or "<" in text:
                template += term_map + " " + text + ";\n"+indentation[0:-1]+"];\n"
            else:
                template += term_map + " \"" + text + "\";\n" + indentation[0:-1] + "];\n"

    return template


def generate_rml_template(yarrrml_template):
    references = 0
    for i in range(len(yarrrml_template)):
        if yarrrml_template[i]=="$" and yarrrml_template[i+1]=="(":
            references = references + 1
    yarrrml_template = yarrrml_template.replace("$(", "{")
    yarrrml_template = "}".join(yarrrml_template.rsplit(")", references))
    return yarrrml_template





import sys
from ruamel.yaml import YAML
import argparse
from rdflib import Graph
from . import translate, inverse_translation, merge_mappings
from .constants import *


def write_results(args, mapping):
    if type(mapping) is str:
        output_file = open(args.output_mapping_path, "w")
        output_file.write(mapping)
        output_file.close()
    elif type(mapping) is dict:
        with open(args.output_mapping_path, "wb") as f:
            yaml = YAML()
            yaml.width = 3000
            yaml.default_flow_style = False
            yaml.dump(mapping, f)


def parse_inputs(args):
    input_format = RML_URI
    yaml = YAML(typ='safe', pure=True)
    if args.input_mapping_path and args.output_mapping_path:
        if args.input_mapping_path.endswith('.yml') or args.input_mapping_path.endswith('.yaml'):
            with open(args.input_mapping_path) as f:
                input_data = yaml.load(f)
        elif args.input_mapping_path.endswith('.ttl') or args.input_mapping_path.endswith(
                '.rml') or args.input_mapping_path.endswith('.r2rml'):
            input_data = Graph()
            input_data.parse(args.input_mapping_path, format="turtle")
        else:
            sys.tracebacklimit = 0
            logger.error("Input mapping does not have a valid extension (.yml, .yaml, .ttl, .rml, or .r2rml)")
            raise Exception("Change your mapping extension to a valid one")

        if args.format is not None and args.format == "R2RML":
            input_format = R2RML_URI
    elif args.merge_mapping:
        input_data = []
        for mapping in args.merge_mapping:
            with open(mapping) as f:
                input_data.append(yaml.load(f))
    else:
        sys.tracebacklimit = 0
        logger.error("No correct arguments, run python3 -m yatter -h to see the help)")
        raise Exception("")

    return input_format, input_data


def define_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input_mapping_path", required=False, help="Input mapping path in YARRRML or RML")
    parser.add_argument("-o", "--output_mapping_path", required=False, help="Output path for the generated mapping")
    parser.add_argument("-m", "--merge_mapping", nargs="*", required=False, help="List of mappings to be merged")
    parser.add_argument("-f", "--format", required=False,
                        help="Mapping input format. Values: R2RML or RML (RML by default)")
    return parser


def main():
    args = define_args().parse_args()
    mapping_format, mapping_data = parse_inputs(args)
    if type(mapping_data) is Graph:
        mapping_content = inverse_translation(mapping_data, mapping_format)
    elif args.merge_mapping:
        mapping_content = merge_mappings(mapping_data)
    else:
        mapping_content = translate(mapping_data, mapping_format)
    write_results(args, mapping_content)





from .constants import *


def add_logical_targets(yarrrml_data, external_targets):
    logical_targets = []
    for target in external_targets:
        logical_targets.extend(generate_logical_target(external_targets.get(target), target))
    for mapping in yarrrml_data.get(YARRRML_MAPPINGS):
        mapping_data = yarrrml_data.get(YARRRML_MAPPINGS).get(mapping)
        add_internal_logical_target(mapping, mapping_data, logical_targets, external_targets)
    return logical_targets


def add_internal_logical_target(mapping_id, mapping_data, internal_targets, external_targets, local_target_id=0):
    keys = mapping_data.keys()
    for key in keys:
        if type(mapping_data[key]) is list:
            for i in range(len(mapping_data[key])):
                value = mapping_data[key][i]
                if YARRRML_TARGETS in value:
                    target_value = value[YARRRML_TARGETS]
                    if type(target_value) is dict or type(target_value) is list:
                        external_targets_keys = [external_targets.get(external_target) for external_target in external_targets]
                        for target_yaml_data in target_value:
                            if YARRRML_ACCESS not in target_yaml_data:
                                target_id, target_yaml_data = target_yaml_data.popitem()
                            if not target_yaml_data in external_targets_keys:
                                logical_target_id = "logical_target_" + mapping_id + "_" + str(local_target_id)
                                internal_targets.extend(generate_logical_target(target_yaml_data, logical_target_id))
                                mapping_data[key][i][YARRRML_TARGETS] = logical_target_id
                                local_target_id += 1
                            else:
                                mapping_data[key][i][YARRRML_TARGETS] = target_id
                else:
                    if type(value) is list:
                        for v in value:
                            if type(v) is dict:
                                add_internal_logical_target(mapping_id, v, internal_targets, external_targets, local_target_id)
                    elif type(value) is dict:
                        add_internal_logical_target(mapping_id, value, internal_targets, external_targets, local_target_id)
        elif YARRRML_TARGETS in mapping_data[key]:
            target_value = mapping_data[key][YARRRML_TARGETS]
            logical_target_id = "logical_target_" + mapping_id + "_" + str(local_target_id)
            internal_targets.extend(generate_logical_target(target_value, logical_target_id))
            mapping_data[key][YARRRML_TARGETS] = logical_target_id
            local_target_id += 1


def generate_logical_target(target_yaml_data, id_target):
    logical_targets = []
    output_type = None
    format = None
    compression = None
    target_yaml_data_list = target_yaml_data
    if isinstance(target_yaml_data_list, dict):
        target_yaml_data_list = [target_yaml_data]
    for target_yaml_data in target_yaml_data_list:
        logical_target = ["<" + id_target + "> a " + RML_LOGICAL_TARGET_CLASS + ";\n"]

        access = target_yaml_data[YARRRML_ACCESS]
        if YARRRML_TYPE in target_yaml_data:
            output_type = target_yaml_data[YARRRML_TYPE]
        if YARRRML_SERIALIZATION in target_yaml_data:
            format = "formats:" + YARRRML_OUTPUT_FORMAT[target_yaml_data[YARRRML_SERIALIZATION]]
        if YARRRML_COMPRESSION in target_yaml_data:
            compression = "comp:" + target_yaml_data[YARRRML_COMPRESSION]

        logical_target.append("\t " + RML_TARGET + " [\n\t\t")

        if output_type == "sparql":
            logical_target.append("sd:endpoint <" + access + ">;\n\t\tsd:supportedLanguage sd:SPARQL11Update\n\t];\n")
        elif output_type == "dcat":
            logical_target.append("a dcat:Dataset;\n\t\tdcat:dataDump <" + access + ">\n\t];\n")
        else:
            logical_target.append("a void:Dataset;\n\t\tvoid:dataDump <" + access + ">\n\t];\n")

        if format:
            logical_target.append("\t " + RML_SERIALIZATION + " " + format + ";")
        else:
            logical_target.append("\t " + RML_SERIALIZATION + " formats:N-Quads;")

        if compression:
            logical_target.append("\n\t " + RML_COMPRESSION + " " + compression + ";")
        logical_targets.append("".join(logical_target)[0:-1] + ".\n\n")
    return logical_targets